"""
Writer for the HDF5_ file format.

.. attention::
   requires that the h5py_ package is installed.

.. _HDF5: https://www.hdfgroup.org/
.. _h5py: https://www.h5py.org/
"""
import os

import numpy as np

try:
    import h5py
except ImportError:
    h5py = None

from ..base import Root
from ..base import Writer
from ..constants import IS_PYTHON2
from ..metadata import Metadata
from ..utils import is_file_readable

if not IS_PYTHON2:
    unicode = str


class HDF5Writer(Writer):
    """Create a HDF5_ writer.

    You can use :class:`HDF5Writer` as a :ref:`context manager <with>`.
    For example,

    .. code-block:: python

        with HDF5Writer('my_file.h5') as root:
            root.create_dataset('dset', data=[1, 2, 3])

    This will automatically write `root` to the specified file when
    the :ref:`with <with>` block exits.
    """

    def write(self, file=None, root=None, **kwargs):
        """Write to a HDF5_ file.

        Parameters
        ----------
        file : :term:`path-like <path-like object>` or :term:`file-like <file object>`, optional
            The file to write the `root` to. If :data:`None` then uses the value of
            `file` that was specified when :class:`HDF5Writer` was instantiated.
        root : :class:`~msl.io.base.Root`, optional
            Write `root` in HDF5_ format. If :data:`None` then write the
            :class:`~msl.io.group.Group`\\s and :class:`~msl.io.dataset.Dataset`\\s
            in this :class:`HDF5Writer`.
        **kwargs
            All key-value pairs are passed to :class:`~h5py.File`.
        """
        if h5py is None:
            raise ImportError(
                'You must install h5py to write HDF5 files, run\n'
                '  pip install h5py'
            )

        if file is None:
            file = self.file
        if not file:
            raise ValueError('You must specify a file to write the root to')

        if root is None:
            root = self
        elif not isinstance(root, Root):
            raise TypeError('The root parameter must be a Root object')

        if 'mode' not in kwargs:
            kwargs['mode'] = 'x'  # Create file, fail if exists

        def check_ndarray_dtype(obj):
            if not isinstance(obj, np.ndarray):
                return obj

            # h5py variable-length string
            vstr = h5py.special_dtype(vlen=str)

            if obj.dtype.names is not None:
                convert, dtype = False, []
                for n in obj.dtype.names:
                    typ = obj.dtype.fields[n][0]
                    if isinstance(obj[n].item(0), (unicode, str)):
                        dtype.append((n, vstr))
                        convert = True
                    else:
                        dtype.append((n, typ))
                if convert:
                    return obj.astype(dtype=dtype)
                return obj
            elif obj.dtype.char == 'U':
                return obj.astype(dtype=vstr)
            elif obj.dtype.char == 'O':
                has_complex = False
                for item in obj.flat:
                    if isinstance(item, (unicode, str)):
                        return obj.astype(dtype='S')
                    elif isinstance(item, np.complexfloating):
                        has_complex = True
                    elif item is None:
                        return obj  # let h5py raise the error that HDF5 does not support NULL
                if has_complex:
                    return obj.astype(dtype=complex)
                return obj.astype(dtype=float)
            else:
                return obj

        def meta_to_dict(metadata):
            return dict((k, meta_to_dict(v) if isinstance(v, Metadata) else check_ndarray_dtype(v))
                        for k, v in metadata.items())

        def h5_open(name):
            with h5py.File(name, **kwargs) as h5:
                h5.attrs.update(**meta_to_dict(root.metadata))
                for name, value in root.items():
                    if self.is_dataset(value):
                        try:
                            vertex = h5.create_dataset(name, data=value.data)
                        except TypeError:
                            vertex = h5.create_dataset(name, data=check_ndarray_dtype(value.data))
                    else:
                        vertex = h5.create_group(name)
                    vertex.attrs.update(**meta_to_dict(value.metadata))

        # Calling h5py.File to write to a file on a mapped drive could raise
        # an OSError. This occurred when a local folder was shared and then
        # mapped on the same computer. Opening the file using open() and then
        # passing in the file handle to h5py.File is more universal
        if hasattr(file, 'write'):  # already a file-like object
            h5_open(file)
        else:
            m = kwargs['mode']
            if m in ['x', 'w-']:
                if os.path.isfile(file) or is_file_readable(file):
                    raise OSError(
                        "File exists {!r}\n"
                        "Specify mode='w' if you want to overwrite it.".format(file)
                    )
            elif m == 'r+':
                if not (os.path.isfile(file) or is_file_readable(file)):
                    raise OSError('File does not exist {!r}'.format(file))
            elif m not in ['w', 'a']:
                raise ValueError('Invalid mode {!r}'.format(m))

            with open(file, mode='w+b') as fp:
                h5_open(fp)



try:
    # this try..except block fixes:
    #   DeprecationWarning: Using or importing the ABCs from 'collections' instead
    #   of from 'collections.abc' is deprecated, and in 3.8 it will stop working
    from collections.abc import MutableMapping, KeysView, ItemsView, ValuesView
except ImportError:
    from collections import MutableMapping, KeysView, ItemsView, ValuesView

import pytest

from msl.io.metadata import Metadata


def test_behaves_like_dict():

    meta = Metadata(read_only=False, vertex_name='', x=1)
    assert isinstance(meta, MutableMapping)
    assert len(meta) == 1
    assert meta['x'] == 1
    assert meta.get('x') == 1
    assert meta.get('y', -99) == -99
    assert 'x' in meta
    assert 'y' not in meta

    # test copy()
    cp = meta.copy()
    assert isinstance(cp, Metadata)
    assert id(cp) != id(meta)
    assert not cp.read_only
    assert cp._vertex_name == meta._vertex_name
    assert cp['x'] == 1
    cp['x'] = 2
    assert cp['x'] == 2
    assert meta['x'] == 1
    assert len(cp) == 1

    # test fromkeys, with default=None, and del
    d = meta.fromkeys({'a', 'b', 'c'})
    assert isinstance(d, Metadata)
    assert id(d) != id(meta)
    assert not d.read_only
    assert len(d) == 3
    assert d['a'] is None
    assert d['b'] is None
    assert d['c'] is None
    del d['a']
    assert 'a' not in meta
    with pytest.raises(KeyError):
        _ = d['a']
    del d['c']
    with pytest.raises(KeyError):
        _ = d['c']
    with pytest.raises(AttributeError):
        del d.c
    assert len(d) == 1

    # test fromkeys(), with default != None, and clear()
    d = meta.fromkeys({'a', 'b', 'c', 'd'}, value='foo')
    assert isinstance(d, Metadata)
    assert id(d) != id(meta)
    assert not d.read_only
    assert len(d) == 4
    assert d['a'] == 'foo'
    assert d['b'] == 'foo'
    assert d['c'] == 'foo'
    assert d['d'] == 'foo'
    assert 'a' not in meta
    d.clear()
    assert len(d) == 0

    # test pop()
    assert 'x' in meta
    assert meta.pop('x') == 1
    assert 'x' not in meta
    with pytest.raises(KeyError):
        meta.pop('x')
    assert meta.pop('x', 22) == 22

    # test popitem()
    assert len(meta) == 0
    meta['x'] = 1
    assert len(meta) == 1
    key, value = meta.popitem()
    assert key == 'x'
    assert value == 1
    assert len(meta) == 0

    # test setdefault()
    assert 'foo' not in meta
    meta.setdefault('xxx')
    assert meta['xxx'] is None
    meta.setdefault('foo', 'bar')
    assert meta['foo'] == 'bar'

    # test update()
    assert len(meta) == 2
    meta.update(dict(a=1, b=2, c=3, hello='world!'))
    assert len(meta) == 6
    assert meta['xxx'] is None
    assert meta['foo'] == 'bar'
    assert meta['a'] == 1
    assert meta['b'] == 2
    assert meta['c'] == 3
    assert meta['hello'] == 'world!'

    expected_keys = ['xxx', 'foo', 'a', 'b', 'c', 'hello']
    expected_values = [None, 'bar', 1, 2, 3, 'world!']

    # test items()
    items = meta.items()
    assert isinstance(items, ItemsView)
    keys = [k for k in expected_keys]
    values = [v for v in expected_values]
    count = 0
    for key, value in items:
        assert key in keys
        keys.pop(keys.index(key))
        assert value in values
        values.pop(values.index(value))
        count += 1
    assert count == 6

    # test keys()
    keys = meta.keys()
    assert isinstance(keys, KeysView)
    keys_ = [k for k in expected_keys]
    count = 0
    for key in keys:
        assert key in keys_
        keys_.pop(keys_.index(key))
        count += 1
    assert count == 6

    # test values()
    values = meta.values()
    assert isinstance(values, ValuesView)
    values_ = [v for v in expected_values]
    count = 0
    for value in values:
        assert value in values_
        values_.pop(values_.index(value))
        count += 1
    assert count == 6

    # test iter()
    count = 0
    keys_ = [k for k in expected_keys]
    for key in iter(meta):
        assert isinstance(key, str)
        assert key in keys_
        keys_.pop(keys_.index(key))
        count += 1
    assert count == 6


def test_attribute_access():

    meta = Metadata(read_only=True, vertex_name='', x=1, FOO='BaR')
    assert meta['x'] == 1
    assert meta.x == 1
    assert meta['FOO'] == 'BaR'
    assert meta.FOO == 'BaR'

    with pytest.raises(AttributeError):
        _ = meta.something

    with pytest.raises(KeyError):
        _ = meta['something']


def test_nested_dict_as_value():
    meta = Metadata(read_only=True, vertex_name='', none=None, nested={'dict1': {'dict2': {'dict3': (1, 2, 3)}}})
    assert meta['none'] is None
    assert meta.none is None
    assert meta['nested']['dict1']['dict2']['dict3'] == (1, 2, 3)
    assert meta.nested.dict1.dict2.dict3 == (1, 2, 3)
    assert meta.nested['dict1'].dict2['dict3'] == (1, 2, 3)

    with pytest.raises(AttributeError):
        _ = meta.none.read_only  # meta.none is not a Metadata object
    with pytest.raises(AttributeError):
        _ = meta['none'].read_only  # meta['none'] is not a Metadata object

    # test that all Metadata objects have the same read-only mode
    assert meta.read_only
    assert meta.nested.read_only
    assert meta.nested.dict1.read_only
    assert meta.nested.dict1.dict2.read_only

    # test read-only mode propagates
    meta.read_only = False
    assert not meta.read_only
    assert not meta.nested.read_only
    assert not meta.nested.dict1.read_only
    assert not meta.nested.dict1.dict2.read_only


def test_read_only():

    meta = Metadata(read_only=True, vertex_name='', foo='bar')
    assert meta.read_only

    # cannot modify an existing value (key access)
    with pytest.raises(ValueError):
        meta['foo'] = 1

    # cannot modify an existing value (attrib access)
    with pytest.raises(ValueError):
        meta.foo = 1

    # cannot delete a key (key access)
    with pytest.raises(ValueError):
        del meta['foo']

    # cannot delete a key (attrib access)
    with pytest.raises(ValueError):
        del meta.foo

    # cannot create a new key-value pair (key access)
    with pytest.raises(ValueError):
        meta['anything'] = -1

    # cannot create a new key-value pair (attrib access)
    with pytest.raises(ValueError):
        meta.anything = -1

    # cannot do an update()
    with pytest.raises(ValueError):
        meta.update({'whatever': 1.234})

    # cannot do a pop()
    with pytest.raises(ValueError):
        meta.pop('foo')

    # cannot do a popitem()
    with pytest.raises(ValueError):
        meta.popitem()

    # cannot do a setdefault(), default value
    with pytest.raises(ValueError):
        meta.setdefault('new_key')

    # cannot do a setdefault(), custom value
    with pytest.raises(ValueError):
        meta.setdefault('new_key', 'new_value')

    # cannot do a clear()
    with pytest.raises(ValueError):
        meta.clear()

    # make sure that meta has the expected data
    assert len(meta) == 1
    assert meta['foo'] == 'bar'

    # make it writeable
    meta.read_only = False
    assert not meta.read_only

    # can modify an existing value (key access)
    meta['foo'] = 1

    # can modify an existing value (attrib access)
    assert meta.foo == 1
    meta.foo = 2

    # can create a new key-value pair (key access)
    meta['negative'] = -1

    # can create a new key-value pair (attrib access)
    meta.one_thousand = 1000

    # can do an update()
    meta.update({'whatever': 1.234})

    # can do a setdefault(), default value
    meta.setdefault('none_value')

    # can do a setdefault(), custom value
    meta.setdefault('custom', 'new_value')

    # make sure that meta has the expected data
    assert len(meta) == 6
    assert meta['foo'] == 2
    assert meta.negative == -1
    assert meta['one_thousand'] == 1000
    assert meta.whatever == 1.234
    assert meta['none_value'] is None
    assert meta.custom == 'new_value'

    # can delete a key (key access)
    del meta['foo']
    del meta['custom']

    # can delete a key (attrib access)
    del meta.one_thousand
    del meta.none_value

    # can do a pop()
    value = meta.pop('negative')
    assert value == -1

    # can do a popitem()
    assert len(meta) == 1
    key, value = meta.popitem()
    assert key == 'whatever'
    assert value == 1.234

    # can do a clear()
    meta.update({'a': 'hello', 'b': 'world'})
    assert len(meta) == 2
    meta.clear()
    assert len(meta) == 0

    # make it read only again
    meta.read_only = True
    assert meta.read_only

    # cannot create a new key-value pair (key access)
    with pytest.raises(ValueError):
        meta['anything'] = -1

    # cannot create a new key-value pair (attrib access)
    with pytest.raises(ValueError):
        meta.anything = -1



"""
A :class:`Dataset` is essentially a :class:`numpy.ndarray` with :class:`~msl.io.metadata.Metadata`.
"""
import numpy as np

from .vertex import Vertex


class Dataset(Vertex):

    def __init__(self, name, parent, read_only, shape=(0,), dtype=float, buffer=None,
                 offset=0, strides=None, order=None, data=None, **metadata):
        """A :class:`Dataset` is essentially a :class:`numpy.ndarray` with :class:`~msl.io.metadata.Metadata`.

        Do not instantiate directly. Create a new :class:`Dataset` using
        :meth:`~msl.io.group.Group.create_dataset`.

        Parameters
        ----------
        name : :class:`str`
            A name to associate with this :class:`Dataset`.
        parent : :class:`~msl.io.group.Group`
            The parent :class:`~msl.io.group.Group` to the :class:`Dataset`.
        read_only : :class:`bool`
            Whether the :class:`Dataset` is to be accessed in read-only mode.
        shape
            See :class:`numpy.ndarray`
        dtype
            See :class:`numpy.ndarray`
        buffer
            See :class:`numpy.ndarray`
        offset
            See :class:`numpy.ndarray`
        strides
            See :class:`numpy.ndarray`
        order
            See :class:`numpy.ndarray`
        data
            If not :data:`None` then it must be either a :class:`numpy.ndarray` or
            an array-like object which will be passed to :func:`numpy.asarray`,
            as well as `dtype` and `order`, to be used as the underlying data.
        **metadata
            All other key-value pairs will be used as
            :class:`~msl.io.metadata.Metadata` for this :class:`Dataset`.
        """
        super(Dataset, self).__init__(name, parent, read_only, **metadata)

        if data is None:
            self._data = np.ndarray(
                shape, dtype=dtype, buffer=buffer, offset=offset, strides=strides, order=order
            )
        else:
            if isinstance(data, np.ndarray):
                self._data = data
            else:
                self._data = np.asarray(data, dtype=dtype, order=order)

        self.read_only = read_only

    def __repr__(self):
        return '<{} {!r} shape={} dtype={!r} ({} metadata)>'\
            .format(self.__class__.__name__, self._name, self._data.shape, self._data.dtype.str, len(self.metadata))

    def __str__(self):
        return repr(self._data)

    @property
    def read_only(self):
        """:class:`bool`: Whether this :class:`Dataset` is in read-only mode.

        This is equivalent to setting the ``WRITEABLE`` property in :meth:`numpy.ndarray.setflags`.
        """
        return not self._data.flags.writeable

    @read_only.setter
    def read_only(self, value):
        val = bool(value)
        self._metadata.read_only = val
        self._data.setflags(write=not val)

    def copy(self, read_only=None):
        """Create a copy of this :class:`Dataset`.

        Parameters
        ----------
        read_only : :class:`bool`, optional
            Whether the copy should be created in read-only mode. If :data:`None` then
            creates a copy using the mode for the :class:`Dataset` that is being copied.

        Returns
        -------
        :class:`Dataset`
            A copy of this :class:`Dataset`.
        """
        return Dataset(
            self._name,
            self._parent,
            self._read_only if read_only is None else read_only,
            data=self._data.copy(),
            **self._metadata.copy()
        )

    @property
    def data(self):
        """:class:`numpy.ndarray`: The data of the :class:`Dataset`.

        .. note::
           You do not have to call this attribute to get access to the
           :class:`numpy.ndarray`. You can directly call the
           :class:`numpy.ndarray` attribute from the :class:`Dataset`
           object.

           For example,

           .. invisible-code-block: pycon

              >>> SKIP_IF_PYTHON_LESS_THAN_36()
              >>> from msl.io import JSONWriter
              >>> root = JSONWriter()
              >>> dset = root.create_dataset('my_data', data=[[0, 1, 2], [3, 4, 5], [6, 7, 8], [9, 10, 11]])

           .. code-block:: pycon

              >>> dset
              <Dataset '/my_data' shape=(4, 3) dtype='<f8' (0 metadata)>
              >>> dset.data
              array([[ 0.,  1.,  2.],
                     [ 3.,  4.,  5.],
                     [ 6.,  7.,  8.],
                     [ 9., 10., 11.]])
              >>> dset.size
              12
              >>> dset.tolist()
              [[0.0, 1.0, 2.0], [3.0, 4.0, 5.0], [6.0, 7.0, 8.0], [9.0, 10.0, 11.0]]
              >>> dset.mean(axis=0)
              array([4.5, 5.5, 6.5])
              >>> dset[::2]
              array([[0., 1., 2.],
                     [6., 7., 8.]])

        """
        return self._data

    def __getitem__(self, item):
        return self._data[item]

    def __setitem__(self, item, value):
        self._data[item] = value

    def __getattr__(self, item):
        try:
            return getattr(self._data, item)
        except AttributeError as err:
            try:
                return self._data[item]
            except (IndexError, ValueError):
                pass
            raise err

    def __len__(self):
        # if the ndarray is a scalar then the following exception is raised
        #   TypeError: len() of unsized object
        try:
            return len(self._data)
        except TypeError:
            return self._data.size

    def __iter__(self):
        return iter(self._data)

    def __add__(self, rhs):
        return self._data + rhs

    def __radd__(self, lhs):
        return lhs + self._data

    def __sub__(self, rhs):
        return self._data - rhs

    def __rsub__(self, lhs):
        return lhs - self._data

    def __mul__(self, rhs):
        return self._data * rhs

    def __rmul__(self, lhs):
        return lhs * self._data

    def __matmul__(self, rhs):
        # TODO could use the @ operator when dropping support for Python 2
        return np.matmul(self._data, rhs)

    def __rmatmul__(self, lhs):
        # TODO could use the @ operator when dropping support for Python 2
        return np.matmul(lhs, self._data)

    def __truediv__(self, rhs):
        return self._data / rhs

    def __rtruediv__(self, lhs):
        return lhs / self._data

    def __div__(self, rhs):
        # TODO could remove this method when dropping support for Python 2
        return self._data / rhs

    def __rdiv__(self, lhs):
        # TODO could remove this method when dropping support for Python 2
        return lhs / self._data

    def __floordiv__(self, rhs):
        return self._data // rhs

    def __rfloordiv__(self, lhs):
        return lhs // self._data

    def __mod__(self, rhs):
        return self._data % rhs

    def __rmod__(self, lhs):
        return lhs % self._data

    def __divmod__(self, rhs):
        return np.divmod(self._data, rhs)

    def __rdivmod__(self, lhs):
        return np.divmod(lhs, self._data)

    def __pow__(self, rhs):
        return self._data ** rhs

    def __rpow__(self, lhs):
        return lhs ** self._data

    def __lshift__(self, rhs):
        return self._data << rhs

    def __rlshift__(self, lhs):
        return lhs << self._data

    def __rshift__(self, rhs):
        return self._data >> rhs

    def __rrshift__(self, lhs):
        return lhs >> self._data

    def __and__(self, rhs):
        return self._data & rhs

    def __rand__(self, lhs):
        return lhs & self._data

    def __xor__(self, rhs):
        return self._data ^ rhs

    def __rxor__(self, lhs):
        return lhs ^ self._data

    def __or__(self, rhs):
        return self._data | rhs

    def __ror__(self, lhs):
        return lhs | self._data

    def __neg__(self):
        return -self._data

    def __pos__(self):
        return self._data

    def __abs__(self):
        return abs(self._data)

    def __invert__(self):
        return ~self._data



"""
Read a file that was created by :class:`~msl.io.writers.json_.JSONWriter`.
"""
import codecs
import json
from io import BufferedIOBase

import numpy as np

from ..base import Reader
from ..constants import IS_PYTHON2
from ..utils import register


@register
class JSONReader(Reader):
    """Read a file that was created by :class:`~msl.io.writers.json_.JSONWriter`."""

    @staticmethod
    def can_read(file, **kwargs):
        """Checks if the text ``MSL JSONWriter`` is in the first line of the file."""
        if isinstance(file, BufferedIOBase):
            text = Reader.get_bytes(file, 21, 34).decode()
        else:
            text = Reader.get_lines(file, 1, **kwargs)[0][20:34]
        return text == 'MSL JSONWriter'

    def read(self, **kwargs):
        """Read the file that was created by :class:`~msl.io.writers.json_.JSONWriter`

        If a :class:`~msl.io.metadata.Metadata` `key` has a `value` that is a
        :class:`list` then the list is converted to an :class:`~numpy.ndarray`
        with :class:`~numpy.dtype` = :class:`object`

        Parameters
        ----------
        **kwargs
            Accepts `encoding` and `errors` keyword arguments which are passed to
            :func:`open`. The default `encoding` value is ``'utf-8'`` and the default
            `errors` value is ``'strict'``. All additional keyword arguments are passed to
            `json.loads <https://docs.python.org/3/library/json.html#json.loads>`_.
        """
        open_kwargs = {
            'encoding': kwargs.get('encoding', 'utf-8'),
            'errors': kwargs.pop('errors', 'strict'),
        }

        if IS_PYTHON2:
            opener = codecs.open  # this allows the encoding and errors kwargs to be used
        else:
            opener = open

        if hasattr(self.file, 'read'):
            self.file.readline()  # skip the first line
            data = self.file.read()
            if isinstance(data, bytes):
                data = data.decode(**open_kwargs)
            dict_ = json.loads(data, **kwargs)
        else:
            with opener(self.file, mode='r', **open_kwargs) as fp:
                fp.readline()  # skip the first line
                dict_ = json.loads(fp.read(), **kwargs)

        def list_to_ndarray(list_):
            # convert a Metadata value to ndarray because one can easily make ndarray read only
            # use dtype=object because it guarantees that the data types are preserved
            # for example,
            #   >>> a = np.asarray([True, -5, 0.002345, 'something', 49.1871524])
            #   >>> a
            #   array(['True', '-5', '0.002345', 'something', '49.1871524'], dtype='<U32')
            # would cast every element to a string
            # also a regular Python list stores items as objects
            return np.asarray(list_, dtype=object)

        def create_group(parent, name, vertex):
            group = self if parent is None else parent.create_group(name)
            for key, value in vertex.items():
                if not isinstance(value, dict):  # Metadata
                    if isinstance(value, list):
                        value = list_to_ndarray(value)
                    group.metadata[key] = value
                elif 'dtype' in value and 'data' in value:  # Dataset
                    kws = dict()
                    for dkey, dval in value.items():
                        if dkey == 'data':
                            pass  # handled in the 'dtype' check
                        elif dkey == 'dtype':
                            if isinstance(dval, list):
                                kws['data'] = np.asarray(
                                    [tuple(row) for row in value['data']],
                                    dtype=[tuple(item) for item in dval])
                            else:
                                kws['data'] = np.asarray(value['data'], dtype=dval)
                        else:  # Metadata
                            if isinstance(dval, list):
                                dval = list_to_ndarray(dval)
                            kws[dkey] = dval
                    group.create_dataset(key, **kws)
                else:  # use recursion to create a sub-Group
                    create_group(group, key, value)

        # create the root group
        create_group(None, '', dict_)



"""
When a new Writer is created it should be added to the list of writers to test.
"""
import os
import tempfile
from io import BufferedWriter
from io import BytesIO
from io import StringIO
from io import TextIOWrapper

import numpy as np

try:
    import h5py
except ImportError:
    h5py = None

from msl.io import HDF5Writer
from msl.io import JSONWriter
from msl.io import read
from msl.io.constants import IS_PYTHON2

# Append new Writers to test
writers = [JSONWriter, HDF5Writer]


def fill_root_with_data(root):
    root.add_metadata(x=1, foo='bar')
    b = root.create_group('a/b')
    b.create_dataset('points', data=[[1, 2], [3, 4], [-5, 6], [7, -8]], cartesian=True)
    root.a.add_metadata(two=2.0)


def assert_root_data(root):
    assert len(list(root.groups())) == 2
    assert len(list(root.descendants())) == 2
    assert len(list(root.ancestors())) == 0
    assert len(list(root.datasets())) == 1
    assert len(root.metadata) == 2
    assert root.metadata['x'] == 1
    assert root.metadata.foo == 'bar'
    assert 'a' in root
    assert 'b' in root.a
    assert 'points' in root.a.b
    assert len(root.a.metadata) == 1
    assert root.a.metadata.two == 2.0
    points = root.a.b.points
    assert len(points.metadata) == 1
    assert points.metadata.cartesian and isinstance(points.metadata.cartesian, (bool, np.bool_))
    assert points.shape == (4, 2)
    assert isinstance(points[0, 1], float)
    assert np.array_equal(points.max(), 7.0)
    assert np.array_equal(points.min(axis=0), [-5.0, -8.0])
    assert np.array_equal(points.max(axis=1), [2.0, 4.0, 6.0, 7.0])
    assert np.array_equal(points, [[1.0, 2.0], [3.0, 4.0], [-5.0, 6.0], [7.0, -8.0]])


def test_string_io():
    # write Root to a StringIO stream and then read it back
    for writer in writers:
        if writer is HDF5Writer:
            continue  # The HDF5Writer cannot write to a text-based stream

        with StringIO() as buf:
            with writer(buf) as root:
                assert isinstance(root.file, StringIO)
                assert repr(root) == "<{} 'StringIO' (0 groups, 0 datasets, 0 metadata)>".format(writer.__name__)
                fill_root_with_data(root)
                assert_root_data(root)
                assert repr(root) == "<{} 'StringIO' (2 groups, 1 datasets, 2 metadata)>".format(writer.__name__)

            buf.seek(0)
            root2 = read(buf)
            assert isinstance(root2.file, StringIO)
            assert_root_data(root2)


def test_bytes_io():
    # write Root to a BytesIO stream and then read it back
    for writer in writers:
        if writer is HDF5Writer and h5py is None:
            continue
        with BytesIO() as buf:
            with writer(buf) as root:
                assert isinstance(root.file, BytesIO)
                assert repr(root) == "<{} 'BytesIO' (0 groups, 0 datasets, 0 metadata)>".format(writer.__name__)
                fill_root_with_data(root)
                assert_root_data(root)
                assert repr(root) == "<{} 'BytesIO' (2 groups, 1 datasets, 2 metadata)>".format(writer.__name__)

            buf.seek(0)
            root2 = read(buf)
            assert isinstance(root2.file, BytesIO)
            assert_root_data(root2)


def test_open_text():
    # write Root to a text-io stream and then read it back
    path = os.path.join(tempfile.gettempdir(), 'textfile.txt')
    if os.path.isfile(path):
        os.remove(path)

    for writer in writers:
        if writer is HDF5Writer:
            continue  # The HDF5Writer cannot write to a text-based stream

        with open(path, mode='wt') as fp:
            with writer(fp) as root:
                if IS_PYTHON2:
                    assert isinstance(root.file, file)
                else:
                    assert isinstance(root.file, TextIOWrapper)
                assert repr(root) == "<{} 'textfile.txt' (0 groups, 0 datasets, 0 metadata)>".format(writer.__name__)
                fill_root_with_data(root)
                assert_root_data(root)
                assert repr(root) == "<{} 'textfile.txt' (2 groups, 1 datasets, 2 metadata)>".format(writer.__name__)

        root2 = read(path)
        assert root2.file == path
        assert_root_data(root2)

        os.remove(path)


def test_open_binary():
    # write Root to a binary-io stream and then read it back
    path = os.path.join(tempfile.gettempdir(), 'binaryfile.bin')
    if os.path.isfile(path):
        os.remove(path)

    for writer in writers:
        if writer is HDF5Writer and h5py is None:
            continue
        with open(path, mode='wb') as fp:
            with writer(fp) as root:
                if IS_PYTHON2:
                    assert isinstance(root.file, file)
                else:
                    assert isinstance(root.file, BufferedWriter)
                assert repr(root) == "<{} 'binaryfile.bin' (0 groups, 0 datasets, 0 metadata)>".format(writer.__name__)
                fill_root_with_data(root)
                assert_root_data(root)
                assert repr(root) == "<{} 'binaryfile.bin' (2 groups, 1 datasets, 2 metadata)>".format(writer.__name__)

        root2 = read(path)
        assert root2.file == path
        assert_root_data(root2)

        os.remove(path)



import os
from datetime import date
from datetime import datetime

import pytest

from msl.io import ExcelReader
from msl.io.readers._xlrd import Book


def test_raises():

    file = os.path.join(os.path.dirname(__file__), 'samples', 'table.xls')

    # file does not exist
    with pytest.raises((IOError, OSError)):
        ExcelReader('does not exist')

    # more than one sheet in the Excel workbook
    with pytest.raises(ValueError, match=r'You must specify the name of the sheet'):
        ExcelReader(file).read()

    # the sheet does not exist in the Excel workbook
    with pytest.raises(ValueError, match=r'There is no sheet named'):
        ExcelReader(file).read(sheet='XXXYYYZZZ')


def test_on_demand_default():
    file = os.path.join(os.path.dirname(__file__), 'samples', 'table.xlsx')
    excel = ExcelReader(file)
    assert excel.workbook.on_demand is True


@pytest.mark.parametrize('on_demand', [True, False])
def test_cell(on_demand):
    file = os.path.join(os.path.dirname(__file__), 'samples', 'table.xlsx')
    excel = ExcelReader(file, on_demand=on_demand)
    assert excel.workbook.on_demand is on_demand
    values = [
        ('timestamp', 'val1', 'uncert1', 'val2', 'uncert2'),
        (datetime(2019, 9, 11, 14, 6, 55), -0.505382, 0.000077, 0.501073, 0.000079),
        (datetime(2019, 9, 11, 14, 6, 59), -0.505191, 0.000066, 0.500877, 0.000083),
        (datetime(2019, 9, 11, 14, 7, 3), -0.505308, 0.000086, 0.500988, 0.000087),
        (datetime(2019, 9, 11, 14, 7, 7), -0.505250, 0.000119, 0.500923, 0.000120),
        (datetime(2019, 9, 11, 14, 7, 11), -0.505275, 0.000070, 0.500965, 0.000088),
        (datetime(2019, 9, 11, 14, 7, 15), -0.505137, 0.000079, 0.500817, 0.000085),
        (datetime(2019, 9, 11, 14, 7, 19), -0.505073, 0.000099, 0.500786, 0.000084),
        (datetime(2019, 9, 11, 14, 7, 23), -0.505133, 0.000088, 0.500805, 0.000076),
        (datetime(2019, 9, 11, 14, 7, 27), -0.505096, 0.000062, 0.500759, 0.000062),
        (datetime(2019, 9, 11, 14, 7, 31), -0.505072, 0.000142, 0.500739, 0.000149)
    ]

    assert excel.file == file
    assert excel.sheet_names() == ('A1', 'BH11', 'AEX154041')
    assert isinstance(excel.workbook, Book)
    sheet = excel.workbook.sheet_by_name('A1')
    assert sheet.nrows == 11
    assert sheet.ncols == 5
    sheet = excel.workbook.sheet_by_name('BH11')
    assert sheet.nrows == 21
    assert sheet.ncols == 64
    sheet = excel.workbook.sheet_by_name('AEX154041')
    assert sheet.nrows == 154051
    assert sheet.ncols == 834

    # single cell
    assert excel.read(cell='A1', sheet='A1') == 'timestamp'
    assert excel.read(cell='B2', sheet='A1') == -0.505382
    assert excel.read(cell='C7', sheet='A1') == 0.000079
    assert excel.read(cell='A100', sheet='A1') is None  # A100 is empty (also out of bounds)
    assert excel.read(cell='BI12', sheet='BH11') == -0.505382
    assert excel.read(cell='A1', sheet='BH11') is None  # A1 is empty
    assert excel.read(cell='ZZ1000', sheet='BH11') is None  # ZZ1000 is empty (also out of bounds)
    assert excel.read(cell='AFB154045', sheet='AEX154041') == 0.00012
    assert excel.read(cell='AA25', sheet='AEX154041') is None  # AA25 is empty
    assert excel.read(cell='BAA200000', sheet='AEX154041') is None  # BAA200000 is empty (also out of bounds)

    # single row
    assert excel.read(cell='A1:E1', sheet='A1') == [values[0]]
    assert excel.read(cell='BH11:BL11', sheet='BH11') == [values[0]]
    assert excel.read(cell='A2:E2', sheet='A1') == [values[1]]
    assert excel.read(cell='BH12:BL12', sheet='BH11') == [values[1]]
    assert excel.read(cell='A3:E3', sheet='A1') == [values[2]]
    assert excel.read(cell='BH13:BL13', sheet='BH11') == [values[2]]
    assert excel.read(cell='A4:E4', sheet='A1') == [values[3]]
    assert excel.read(cell='BH14:BL14', sheet='BH11') == [values[3]]
    assert excel.read(cell='A5:E5', sheet='A1') == [values[4]]
    assert excel.read(cell='BH15:BL15', sheet='BH11') == [values[4]]
    assert excel.read(cell='A6:E6', sheet='A1') == [values[5]]
    assert excel.read(cell='BH16:BL16', sheet='BH11') == [values[5]]
    assert excel.read(cell='A7:E7', sheet='A1') == [values[6]]
    assert excel.read(cell='BH17:BL17', sheet='BH11') == [values[6]]
    assert excel.read(cell='A8:E8', sheet='A1') == [values[7]]
    assert excel.read(cell='BH18:BL18', sheet='BH11') == [values[7]]
    assert excel.read(cell='A9:E9', sheet='A1') == [values[8]]
    assert excel.read(cell='BH19:BL19', sheet='BH11') == [values[8]]
    assert excel.read(cell='A10:E10', sheet='A1') == [values[9]]
    assert excel.read(cell='BH20:BL20', sheet='BH11') == [values[9]]
    assert excel.read(cell='A11:E11', sheet='A1') == [values[10]]
    assert excel.read(cell='BH21:BL21', sheet='BH11') == [values[10]]
    assert excel.read(cell='A6:B6', sheet='A1') == [values[5][:2]]
    assert excel.read(cell='A12:C12', sheet='A1') == []  # row 12 is empty (also out of bounds)
    assert excel.read(cell='A1000:Z1000', sheet='A1') == []  # row 1000 is empty (also out of bounds)
    assert excel.read(cell='BH22:BL22', sheet='BH11') == []  # row 22 is empty (also out of bounds)
    assert excel.read(cell='A1000:ZZ1000', sheet='BH11') == []  # row 1000 is empty (also out of bounds)
    assert excel.read(cell='A1:A1', sheet='A1') == [('timestamp',)]
    assert excel.read(cell='D9:D9', sheet='A1') == [(0.500805,)]

    # single column
    assert excel.read(cell='A:A', sheet='A1') == [(item[0],) for item in values]
    assert excel.read(cell='B:B', sheet='A1') == [(item[1],) for item in values]
    assert excel.read(cell='C:C', sheet='A1') == [(item[2],) for item in values]
    assert excel.read(cell='D:D', sheet='A1') == [(item[3],) for item in values]
    assert excel.read(cell='E:E', sheet='A1') == [(item[4],) for item in values]
    assert excel.read(cell='F:F', sheet='A1') == []  # column F is empty (also out of bounds)
    assert excel.read(cell='ABC:ABC', sheet='A1') == []  # column ABC is empty (also out of bounds)
    assert excel.read(cell='BH:BH', sheet='BH11') == [(None,) for _ in range(10)] + [(item[0],) for item in values]
    assert excel.read(cell='A:A', sheet='BH11') == [(None,) for _ in range(21)]  # column A is empty
    assert excel.read(cell='BG:BG', sheet='BH11') == [(None,) for _ in range(21)]  # column BG is empty

    # 2D slices
    assert excel.read(cell='A:E', sheet='A1') == values
    assert excel.read(cell='A1:E11', sheet='A1') == values
    assert excel.read(cell='A1:AAA10000', sheet='A1') == values  # slicing out of range is okay
    assert excel.read(cell='A1:C6', sheet='A1') == [row[:3] for row in values[:6]]
    assert excel.read(cell='A10:E11', sheet='A1') == values[-2:]
    assert excel.read(cell='A10:E1000', sheet='A1') == values[-2:]  # slicing out of range is okay
    assert excel.read(cell='A:E', sheet='BH11') == [tuple(None for _ in range(5)) for _ in range(21)]
    new = [tuple(None for _ in range(6))]
    for row in values:
        new.append((None,) + row)
    assert excel.read(cell='BG10:BM22', sheet='BH11') == new
    assert excel.read(cell='BK20:BL21', sheet='BH11') == [row[-2:] for row in values[-2:]]
    assert excel.read(cell='AEX154041:AFB154051', sheet='AEX154041') == values
    assert excel.read(cell='AEX154041:ZZZ1000000', sheet='AEX154041') == values  # slicing out of range is okay
    assert excel.read(cell='AEY154042:AFA154044', sheet='AEX154041') == [row[1:4] for row in values[1:4]]
    assert excel.read(cell='J1:M10', sheet='A1') == []

    # calling close() multiple times is okay
    for _ in range(10):
        excel.close()


@pytest.mark.parametrize('on_demand', [True, False])
def test_datatypes(on_demand):
    # the following workbook only contains 1 sheet, so we don't have to specify the sheet
    file = os.path.join(os.path.dirname(__file__), 'samples', 'excel_datatypes.xlsx')
    excel = ExcelReader(file, on_demand=on_demand)
    assert excel.workbook.on_demand is on_demand
    assert excel.file == file
    assert excel.workbook.nsheets == 1
    assert excel.sheet_names() == ('Sheet1',)
    assert excel.read(cell='A1') == 1.23  # '$1.23'
    assert excel.read(cell='B1') is True
    assert excel.read(cell='C1') == datetime(2019, 9, 13, 13, 20, 22)
    assert excel.read(cell='C1', as_datetime=False) == '2019-09-13 13:20:22'
    assert excel.read(cell='D1') == date(2019, 9, 13)
    assert excel.read(cell='D1', as_datetime=False) == '2019-09-13'
    assert excel.read(cell='A2') == 3.141592653589793
    assert excel.read(cell='B2') == 'some text'
    assert excel.read(cell='C2') is None
    assert excel.read(cell='D2') == 0.34  # '34%'
    assert excel.read(cell='A1:A2') == [(1.23,), (3.141592653589793,)]

    # calling close() multiple times is okay
    for _ in range(10):
        excel.close()



# -*- coding: utf-8 -*-
import pathlib
import socket
import threading
import time
from io import BytesIO

import pytest

try:
    import h5py
except ImportError:
    h5py = None

from helper import read_sample
from helper import roots_equal
from msl.io import JSONWriter
from msl.io import read
from msl.io.readers import JSONReader


def test_raises():
    # file does not exist
    with pytest.raises((IOError, OSError), match=r'No such file'):
        read('does_not.exist')

    # unicode filename
    with pytest.raises((IOError, OSError), match=r'No such file'):
        read(u'Filé döes ñot éxist')

    # no Reader class exists to read this test_read.py file
    with pytest.raises(OSError, match=r'No Reader exists'):
        read(__file__)

    # unicode filename
    with pytest.raises(OSError, match=r'No Reader exists'):
        read_sample(u'uñicödé')


@pytest.mark.skipif(h5py is None, reason='h5py is not installed')
def test_unicode_filename():
    root = read_sample(u'uñicödé.h5')
    assert root.metadata.is_unicode
    assert root.file.endswith(u'uñicödé.h5')
    assert u'café' in root
    assert u'/café' in root
    assert u'café/caña' in root
    assert u'/café/caña' in root
    assert u'caña' in root[u'café']
    assert u'/caña' in root[u'/café']
    assert u'cafécaña' not in root


def test_socket():
    # test that we can read/write a Root object from a socket stream

    # get any available port
    sock = socket.socket()
    sock.bind(('', 0))
    port = sock.getsockname()[1]
    sock.close()

    # the server will run in a separate thread
    def start_server():
        server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server.bind(('localhost', port))
        server.listen(1)
        conn, _ = server.accept()
        with BytesIO(conn.recv(2**18)) as buf:
            # check that the bytes from the client can be read by JSONReader
            assert isinstance(read(buf), JSONReader)
            # send the bytes back to the client
            conn.sendall(buf.getvalue())
        conn.close()

    # start the server
    thread = threading.Thread(target=start_server)
    thread.start()
    time.sleep(1)

    # read the sample file into a BytesIO stream
    client_send_buf = BytesIO()
    with JSONWriter(client_send_buf) as json:
        root_client = read_sample('json_sample.json')
        json.set_root(root_client)

    # send the bytes to the server
    client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    client.connect(('localhost', port))
    client.sendall(client_send_buf.getvalue())
    client_send_buf.close()

    # get the bytes back from the server
    with BytesIO(client.recv(2**18)) as client_recv_buf:
        root_server = read(client_recv_buf)
        assert isinstance(root_server, JSONReader)

    assert roots_equal(root_client, root_server)

    # cleanup
    client.close()
    thread.join()


def test_pathlib():
    root1 = read_sample('json_sample.json')
    root2 = read(pathlib.Path(root1.file))
    assert roots_equal(root1, root2)



"""
Read a Google Sheets spreadsheet.
"""
import os
import re

from .spreadsheet import Spreadsheet
from ..google_api import GCellType
from ..google_api import GDrive
from ..google_api import GSheets

_google_file_id_regex = re.compile(r'^1[a-zA-Z0-9_-]{43}$')


class GSheetsReader(Spreadsheet):

    def __init__(self, file, **kwargs):
        """Read a Google Sheets spreadsheet.

        This class simply provides a convenience for reading information
        from Google spreadsheets. It is not registered as a :class:`~msl.io.base.Reader`
        because the information in a spreadsheet is unstructured and therefore
        one cannot generalize how to parse a spreadsheet to create a
        :class:`~msl.io.base.Root`.

        Parameters
        ----------
        file : :class:`str`
            The ID or path of a Google Sheets spreadsheet.
        **kwargs
            All keyword arguments are passed to :class:`~msl.io.google_api.GSheets`.

        Examples
        --------
        >>> from msl.io import GSheetsReader  # doctest: +SKIP
        >>> sheets = GSheetsReader('Google Drive/registers/equipment.gsheet')  # doctest: +SKIP
        >>> sheets = GSheetsReader('1TI3pM-534SZ5DQTEZ-7HCI04648f8ZpLGbfHWJu9FSo')  # doctest: +SKIP
        """
        super(GSheetsReader, self).__init__(file)

        if not kwargs.get('read_only', True):
            raise ValueError('Must instantiate {} in read-only mode'.format(self.__class__.__name__))

        path, ext = os.path.splitext(file)
        folders, _ = os.path.split(path)
        if ext or folders or not _google_file_id_regex.match(path):
            self._spreadsheet_id = GDrive(**kwargs).file_id(path, mime_type=GSheets.MIME_TYPE)
        else:
            self._spreadsheet_id = path

        self._gsheets = GSheets(**kwargs)
        self._cached_sheet_name = None

    def read(self, cell=None, sheet=None, as_datetime=True):
        """Read values from the Google Sheets spreadsheet.

        Parameters
        ----------
        cell : :class:`str`, optional
            The cell(s) to read. For example, ``C9`` will return a single value
            and ``C9:G20`` will return all values in the specified range. If not
            specified then returns all values in the specified `sheet`.
        sheet : :class:`str`, optional
            The name of the sheet to read the value(s) from. If there is only
            one sheet in the spreadsheet then you do not need to specify the name
            of the sheet.
        as_datetime : :class:`bool`, optional
            Whether dates should be returned as :class:`~datetime.datetime` or
            :class:`~datetime.date` objects. If :data:`False` then dates are
            returned as a string in the format of the spreadsheet cell.

        Returns
        -------
        The value(s) of the requested cell(s).

        Examples
        --------
        .. invisible-code-block: pycon

           >>> SKIP_IF_NO_GOOGLE_SHEETS_READ_TOKEN()
           >>> from msl.io import GSheetsReader
           >>> sheets = GSheetsReader('1TI3pM-534SZ5DQTEZ-7vCI04l48f8ZpLGbfEWJuCFSo', account='testing')

        >>> sheets.read()
        [('temperature', 'humidity'), (20.33, 49.82), (20.23, 46.06), (20.41, 47.06), (20.29, 48.32)]
        >>> sheets.read('B2')
        49.82
        >>> sheets.read('A:A')
        [('temperature',), (20.33,), (20.23,), (20.41,), (20.29,)]
        >>> sheets.read('A1:B1')
        [('temperature', 'humidity')]
        >>> sheets.read('A2:B4')
        [(20.33, 49.82), (20.23, 46.06), (20.41, 47.06)]
        """
        if not sheet:
            if self._cached_sheet_name:
                sheet = self._cached_sheet_name
            else:
                names = self.sheet_names()
                if len(names) != 1:
                    raise ValueError('{!r} contains the following sheets:\n  {}\n'
                                     'You must specify the name of the sheet to read'
                                     .format(self._file, ', '.join(repr(n) for n in names)))
                sheet = names[0]
                self._cached_sheet_name = sheet

        if cell:
            ranges = '{}!{}'.format(sheet, cell)
        else:
            ranges = sheet

        cells = self._gsheets.cells(self._spreadsheet_id, ranges=ranges)

        if sheet not in cells:
            raise ValueError('There is no sheet named {!r} in {!r}'.format(sheet, self._file))

        values = []
        for row in cells[sheet]:
            row_values = []
            for item in row:
                if item.type == GCellType.DATE:
                    value = GSheets.to_datetime(item.value).date() if as_datetime else item.formatted
                elif item.type == GCellType.DATE_TIME:
                    value = GSheets.to_datetime(item.value) if as_datetime else item.formatted
                else:
                    value = item.value
                row_values.append(value)
            values.append(tuple(row_values))

        if not cell:
            return values

        if ':' not in cell:
            if values:
                return values[0][0]
            return

        return values

    def sheet_names(self):
        """Get the names of all sheets in the Google Sheets spreadsheet.

        Returns
        -------
        :class:`tuple` of :class:`str`
            The names of all sheets.
        """
        return self._gsheets.sheet_names(self._spreadsheet_id)



"""
An :class:`~collections.OrderedDict` that can be made read only.
"""
from collections import OrderedDict
try:
    # this try..except block fixes:
    #   DeprecationWarning: Using or importing the ABCs from 'collections' instead
    #   of from 'collections.abc' is deprecated, and in 3.8 it will stop working
    from collections.abc import MutableMapping, KeysView, ItemsView, ValuesView
except ImportError:
    from collections import MutableMapping, KeysView, ItemsView, ValuesView


class Dictionary(MutableMapping):

    def __init__(self, read_only, **kwargs):
        """A :class:`dict` that can be made read only.

        Parameters
        ----------
        read_only : :class:`bool`
            Whether the underlying :class:`dict` should be created in read-only mode.
        **kwargs
            Key-value pairs that are used to create the underlying :class:`dict` object.
        """
        self._read_only = bool(read_only)
        self._mapping = OrderedDict(**kwargs)

    def __repr__(self):
        return '{' + ', '.join('{!r}: {!r}'.format(key, value) for key, value in self._mapping.items()) + '}'

    def __iter__(self):
        return iter(self._mapping)

    def __len__(self):
        return len(self._mapping)

    def __delitem__(self, item):
        self._raise_if_read_only()
        del self._mapping[item]

    def __getitem__(self, item):
        return self._mapping[item]

    def __setitem__(self, item, value):
        self._raise_if_read_only()
        self._mapping[item] = value

    @property
    def read_only(self):
        """:class:`bool`: Whether the underlying :class:`dict` is in read-only mode."""
        return self._read_only

    @read_only.setter
    def read_only(self, value):
        self._read_only = bool(value)

    def clear(self):
        """Remove all items from the dictionary."""
        self._raise_if_read_only()
        self._mapping.clear()

    def keys(self):
        """Return a new view of the dictionary's keys."""
        return KeysView(self)

    def values(self):
        """Return a new view of the dictionary's values."""
        return ValuesView(self)

    def items(self):
        """Return a new view of the dictionary's items, i.e., (key, value) pairs."""
        return ItemsView(self)

    def _raise_if_read_only(self):
        if self._read_only:
            # numpy also raises ValueError if the ndarray is in read-only mode
            raise ValueError('Cannot modify {!r}. It is accessed in read-only mode.'.format(self))

    def _raise_key_error(self, key):
        raise KeyError('{!r} is not in {!r}'.format(key, self))



import os

import numpy as np
import pytest

try:
    import h5py
except ImportError:
    h5py = None

from helper import metadata_equal
from helper import roots_equal
from msl.io import HDF5Writer
from msl.io import JSONWriter
from msl.io import copy
from msl.io import is_dir_accessible
from msl.io import read
from msl.io import read_table

# the Z: drive (if it exists) is a mapped drive to the "Photometry & Radiometry" folder
folder = r'Z:\transfer'

skipif_not_mapped = pytest.mark.skipif(
    not is_dir_accessible(folder),
    reason='mapped drive not available'
)


@skipif_not_mapped
@pytest.mark.skipif(h5py is None, reason='h5py is not installed')
def test_hdf5():
    root = read(os.path.join(os.path.dirname(__file__), 'samples', 'hdf5_sample.h5'))
    w = HDF5Writer(os.path.join(folder, 'msl-io-testing.h5'))
    w.write(root=root, mode='w')
    assert roots_equal(root, read(w.file))
    os.remove(w.file)


@skipif_not_mapped
def test_json():
    root = read(os.path.join(os.path.dirname(__file__), 'samples', 'json_sample.json'))
    w = JSONWriter(os.path.join(folder, 'msl-io-testing.json'))
    w.write(root=root, mode='w')
    assert roots_equal(root, read(w.file))
    os.remove(w.file)


@skipif_not_mapped
def test_copy_and_table():
    tables = {
        'table.csv': ('msl-io-testing.csv', {}),
        'table.txt': ('msl-io-testing.txt', {}),
        'table.xls': ('msl-io-testing.xls', {'sheet': 'A1'}),
    }
    for original, (file, kwargs) in tables.items():
        source = os.path.join(os.path.dirname(__file__), 'samples', original)
        d1 = read_table(source, dtype=object, **kwargs)
        destination1 = os.path.join(folder)
        destination2 = os.path.join(folder, file)
        destination3 = os.path.join(folder, 'msl-io-testing', 'subfolder', file)
        for i, dest in enumerate([destination1, destination2, destination3]):
            dest = copy(source, dest, overwrite=True)
            d2 = read_table(dest, dtype=object, **kwargs)
            if i == 0:
                assert d1.name == d2.name
            assert np.array_equal(d1.data, d2.data)
            assert metadata_equal(d1.metadata, d2.metadata)
            os.remove(dest)
        sub_folder = os.path.dirname(destination3)
        os.rmdir(sub_folder)
        os.rmdir(os.path.dirname(sub_folder))



from helper import read_sample
from msl.io.readers import DRSReader


def test_drs():
    root = read_sample('Lamp_15082018_4.DAT')
    assert isinstance(root, DRSReader)

    assert 'run1' in root
    assert 'scan1' in root['run1']
    assert 'run2' in root
    assert 'scan1' in root['run2']
    assert 'run3' in root
    assert 'scan1' in root['run3']
    assert 'run4' in root
    assert 'scan1' in root['run4']

    assert root['run1'].metadata['wavelength_start'] == 600
    assert root['run2'].metadata['wavelength_start'] == 700
    assert root['run3'].metadata['wavelength_increment'] == 50
    assert root['run4'].metadata['wavelength_end'] == 1000

    assert root.run1.metadata.comment.startswith('Comparison F 637 vs F636')

    assert root.run2.metadata.devices['LAMP 1']['Name'] == 'F636'
    assert root.run2.metadata.devices['LAMP 2']['Name'] == 'F637'
    assert root.run3.metadata.devices['LAMP 1']['Name'] == 'F637'
    assert root.run3.metadata.devices['LAMP 2']['Name'] == 'F636'

    dat = root['run1']['scan1']['dat']
    log636 = root['run1']['scan1']['log-F636']
    log637 = root['run1']['scan1']['log-F637']
    assert root.is_dataset(dat)
    assert root.is_dataset(log636)
    assert root.is_dataset(log637)
    assert dat['F636'][1] == 2.98123609
    assert dat['u(F636)'][1] == 0.00066831
    assert abs(log636.wavelength[0] - 600) < 0.001

    dat = root['run3']['scan1']['dat']
    log636 = root['run3']['scan1']['log-F636']
    log637 = root['run3']['scan1']['log-F637']
    assert root.is_dataset(dat)
    assert root.is_dataset(log636)
    assert root.is_dataset(log637)
    assert dat['F636'][3] == 0.90075543
    assert dat['u(F636)'][3] == 0.00032893
    assert abs(log636.wavelength[4] - 800) < 0.001



"""
Generic class for spreadsheets.
"""
import re
import string

_cell_regex = re.compile(r'^([A-Z]+)(\d*)$')


class Spreadsheet(object):

    def __init__(self, file):
        """Generic class for spreadsheets.

        Parameters
        ----------
        file : :class:`str`
            The location of the spreadsheet on a local hard drive or on a network.
        """
        self._file = file

    @property
    def file(self):
        """:class:`str`: The location of the spreadsheet on a local hard drive or on a network."""
        return self._file

    def read(self, cell=None, sheet=None, as_datetime=True):
        """Read values from the spreadsheet.

        Parameters
        ----------
        cell : :class:`str`, optional
            The cell(s) to read. For example, ``C9`` will return a single value
            and ``C9:G20`` will return all values in the specified range. If not
            specified then returns all values in the specified `sheet`.
        sheet : :class:`str`, optional
            The name of the sheet to read the value(s) from. If there is only
            one sheet in the spreadsheet then you do not need to specify the name
            of the sheet.
        as_datetime : :class:`bool`, optional
            Whether dates should be returned as :class:`~datetime.datetime` or
            :class:`~datetime.date` objects. If :data:`False` then dates are
            returned as a string.

        Returns
        -------
        The value(s) of the requested cell(s).
        """
        raise NotImplementedError

    def sheet_names(self):
        """Get the names of all sheets in the spreadsheet.

        Returns
        -------
        :class:`tuple` of :class:`str`
            The names of all sheets.
        """
        raise NotImplementedError

    @staticmethod
    def to_letters(index):
        """Convert a column index to column letters.

        Parameters
        ----------
        index : :class:`int`
            The column index (zero based).

        Returns
        -------
        :class:`str`
            The corresponding spreadsheet column letter(s).

        Examples
        --------
        .. invisible-code-block: pycon

           >>> from msl.io.readers.spreadsheet import Spreadsheet
           >>> to_letters = Spreadsheet.to_letters

        >>> to_letters(0)
        'A'
        >>> to_letters(1)
        'B'
        >>> to_letters(26)
        'AA'
        >>> to_letters(702)
        'AAA'
        >>> to_letters(494264)
        'ABCDE'
        """
        letters = []
        uppercase = string.ascii_uppercase
        while index >= 0:
            div, mod = divmod(index, 26)
            letters.append(uppercase[mod])
            index = div - 1
        return ''.join(letters[::-1])

    @staticmethod
    def to_indices(cell):
        """Convert a string representation of a cell to row and column indices.

        Parameters
        ----------
        cell : :class:`str`
            The cell. Can be letters only (a column) or letters and a number
            (a column and a row).

        Returns
        -------
        :class:`tuple`
            The (row_index, column_index). If `cell` does not contain a row number
            then the row index is :data:`None`. The row and column index are zero based.

        Examples
        --------
        .. invisible-code-block: pycon

           >>> from msl.io.readers.spreadsheet import Spreadsheet
           >>> to_indices = Spreadsheet.to_indices

        >>> to_indices('A')
        (None, 0)
        >>> to_indices('A1')
        (0, 0)
        >>> to_indices('AA10')
        (9, 26)
        >>> to_indices('AAA111')
        (110, 702)
        >>> to_indices('MSL123456')
        (123455, 9293)
        >>> to_indices('BIPM')
        (None, 41664)
        """
        match = _cell_regex.match(cell)
        if not match:
            raise ValueError('Invalid cell {!r}'.format(cell))

        letters, numbers = match.groups()
        row = max(0, int(numbers) - 1) if numbers else None
        uppercase = string.ascii_uppercase
        col = sum(
            (26**i) * (1+uppercase.index(c))
            for i, c in enumerate(letters[::-1])
        )
        return row, col-1

    @staticmethod
    def to_slices(cells, row_step=None, column_step=None):
        """Convert a range of cells to slices of row and column indices.

        Parameters
        ----------
        cells : :class:`str`
            The cells. Can be letters only (a column) or letters and a number
            (a column and a row).
        row_step : :class:`int`, optional
            The step-by value for the row slice.
        column_step : :class:`int`, optional
            The step-by value for the column slice.

        Returns
        -------
        :class:`slice`
            The row slice.
        :class:`slice`
            The column slice.

        Examples
        --------
        .. invisible-code-block: pycon

           >>> from msl.io.readers.spreadsheet import Spreadsheet
           >>> to_slices = Spreadsheet.to_slices

        >>> to_slices('A:A')
        (slice(0, None, None), slice(0, 1, None))
        >>> to_slices('A:H')
        (slice(0, None, None), slice(0, 8, None))
        >>> to_slices('B2:M10')
        (slice(1, 10, None), slice(1, 13, None))
        >>> to_slices('A5:M100', row_step=2, column_step=4)
        (slice(4, 100, 2), slice(0, 13, 4))
        """
        split = cells.split(':')
        if len(split) != 2:
            raise ValueError('Invalid cell range {!r}'.format(cells))

        r1, c1 = Spreadsheet.to_indices(split[0])
        r2, c2 = Spreadsheet.to_indices(split[1])
        if r1 is None:
            r1 = 0
        if r2 is not None:
            r2 += 1
        if c2 is not None:
            c2 += 1
        return slice(r1, r2, row_step), slice(c1, c2, column_step)



"""
Read an Excel spreadsheet (.xls and .xlsx).
"""
from datetime import datetime

from . import _xlrd
from .spreadsheet import Spreadsheet


class ExcelReader(Spreadsheet):

    def __init__(self, file, **kwargs):
        """Read an Excel spreadsheet (.xls and .xlsx).

        This class simply provides a convenience for reading information
        from Excel spreadsheets. It is not registered as a :class:`~msl.io.base.Reader`
        because the information in an Excel spreadsheet is unstructured and therefore
        one cannot generalize how to parse an Excel spreadsheet to create a
        :class:`~msl.io.base.Root`.

        Parameters
        ----------
        file : :class:`str`
            The location of an Excel spreadsheet on a local hard drive or on a network.
        **kwargs
            All keyword arguments are passed to :func:`~xlrd.open_workbook`. Can use
            an `encoding` keyword argument as an alias for `encoding_override`. The
            default `on_demand` value is :data:`True`.

        Examples
        --------
        >>> from msl.io import ExcelReader  # doctest: +SKIP
        >>> excel = ExcelReader('lab_environment.xlsx')  # doctest: +SKIP
        """
        super(ExcelReader, self).__init__(file)

        # change the default on_demand value
        if 'on_demand' not in kwargs:
            kwargs['on_demand'] = True

        # 'encoding' is an alias for 'encoding_override'
        encoding = kwargs.pop('encoding', None)
        if encoding is not None:
            kwargs['encoding_override'] = encoding

        self._workbook = _xlrd.open_workbook(file, **kwargs)

    @property
    def workbook(self):
        """:class:`~xlrd.book.Book`: The workbook instance."""
        return self._workbook

    def close(self):
        """Calls :meth:`~xlrd.book.Book.release_resources`."""
        self._workbook.release_resources()

    def read(self, cell=None, sheet=None, as_datetime=True):
        """Read values from the Excel spreadsheet.

        Parameters
        ----------
        cell : :class:`str`, optional
            The cell(s) to read. For example, ``C9`` will return a single value
            and ``C9:G20`` will return all values in the specified range. If not
            specified then returns all values in the specified `sheet`.
        sheet : :class:`str`, optional
            The name of the sheet to read the value(s) from. If there is only
            one sheet in the spreadsheet then you do not need to specify the name
            of the sheet.
        as_datetime : :class:`bool`, optional
            Whether dates should be returned as :class:`~datetime.datetime` or
            :class:`~datetime.date` objects. If :data:`False` then dates are
            returned as an ISO 8601 string.

        Returns
        -------
        The value(s) of the requested cell(s).

        Examples
        --------
        .. invisible-code-block: pycon

           >>> from msl.io import ExcelReader
           >>> excel = ExcelReader('./tests/samples/lab_environment.xlsx')

        >>> excel.read()
        [('temperature', 'humidity'), (20.33, 49.82), (20.23, 46.06), (20.41, 47.06), (20.29, 48.32)]
        >>> excel.read('B2')
        49.82
        >>> excel.read('A:A')
        [('temperature',), (20.33,), (20.23,), (20.41,), (20.29,)]
        >>> excel.read('A1:B1')
        [('temperature', 'humidity')]
        >>> excel.read('A2:B4')
        [(20.33, 49.82), (20.23, 46.06), (20.41, 47.06)]
        """
        if not sheet:
            names = self.sheet_names()
            if len(names) > 1:
                raise ValueError('{!r} contains the following sheets:\n  {}\n'
                                 'You must specify the name of the sheet to read'
                                 .format(self._file, ', '.join(repr(n) for n in names)))
            sheet_name = names[0]
        else:
            sheet_name = sheet

        try:
            sheet = self._workbook.sheet_by_name(sheet_name)
        except _xlrd.XLRDError:
            # TODO want to raise ValueError without nested exceptions
            #  Can "raise from None" when dropping Python 2 support
            sheet = None

        if sheet is None:
            raise ValueError('There is no sheet named {!r} in {!r}'.format(sheet_name, self._file))

        if not cell:
            return [tuple(self._value(sheet, r, c, as_datetime) for c in range(sheet.ncols))
                    for r in range(sheet.nrows)]

        split = cell.split(':')
        r1, c1 = self.to_indices(split[0])
        if r1 is None:
            r1 = 0

        if len(split) == 1:
            try:
                return self._value(sheet, r1, c1, as_datetime)
            except IndexError:
                return

        if r1 >= sheet.nrows or c1 >= sheet.ncols:
            return []

        r2, c2 = self.to_indices(split[1])
        r2 = sheet.nrows if r2 is None else min(r2+1, sheet.nrows)
        c2 = min(c2+1, sheet.ncols)
        return [tuple(self._value(sheet, r, c, as_datetime) for c in range(c1, c2))
                for r in range(r1, r2)]

    def sheet_names(self):
        """Get the names of all sheets in the Excel spreadsheet.

        Returns
        -------
        :class:`tuple` of :class:`str`
            The names of all sheets.
        """
        return tuple(self._workbook.sheet_names())

    def _value(self, sheet, row, col, as_datetime):
        """Get the value of a cell."""
        cell = sheet.cell(row, col)
        t = cell.ctype
        if t == _xlrd.XL_CELL_NUMBER:
            if cell.value.is_integer():
                return int(cell.value)
            return cell.value
        elif t == _xlrd.XL_CELL_DATE:
            dt = datetime(*_xlrd.xldate_as_tuple(cell.value, self._workbook.datemode))
            if dt.hour + dt.minute + dt.second + dt.microsecond == 0:
                dt = dt.date()
            if as_datetime:
                return dt
            return str(dt)
        elif t == _xlrd.XL_CELL_BOOLEAN:
            return bool(cell.value)
        elif t == _xlrd.XL_CELL_EMPTY:
            return None
        elif t == _xlrd.XL_CELL_ERROR:
            return _xlrd.error_text_from_code[cell.value]
        else:
            return cell.value.strip()



"""
When a new Writer is created it should be added to the list of writers to test.
"""
import os
import tempfile
from io import StringIO

import numpy as np
import pytest

try:
    import h5py
except ImportError:
    h5py = None

from msl.io import HDF5Writer
from msl.io import JSONWriter
from msl.io import read

# Append new Writers to test
writers = [JSONWriter, HDF5Writer]


def fill_root_with_data(root):
    root.add_metadata(x=1, foo='bar')
    b = root.create_group('a/b')
    b.create_dataset('points', data=[[1, 2], [3, 4], [-5, 6], [7, -8]], cartesian=True)
    root.a.add_metadata(two=2.0)


def assert_root_data(root):
    assert len(list(root.groups())) == 2
    assert len(list(root.descendants())) == 2
    assert len(list(root.ancestors())) == 0
    assert len(list(root.datasets())) == 1
    assert len(root.metadata) == 2
    assert root.metadata['x'] == 1
    assert root.metadata.foo == 'bar'
    assert 'a' in root
    assert 'b' in root.a
    assert 'points' in root.a.b
    assert len(root.a.metadata) == 1
    assert root.a.metadata.two == 2.0
    points = root.a.b.points
    assert len(points.metadata) == 1
    assert points.metadata.cartesian and isinstance(points.metadata.cartesian, (bool, np.bool_))
    assert points.shape == (4, 2)
    assert isinstance(points[0, 1], float)
    assert np.array_equal(points.max(), 7.0)
    assert np.array_equal(points.min(axis=0), [-5.0, -8.0])
    assert np.array_equal(points.max(axis=1), [2.0, 4.0, 6.0, 7.0])
    assert np.array_equal(points, [[1.0, 2.0], [3.0, 4.0], [-5.0, 6.0], [7.0, -8.0]])


def test_none_type():
    for writer in writers:
        if writer is HDF5Writer and h5py is None:
            continue
        with pytest.raises(ValueError) as err:
            with writer() as root:
                assert root.file is None
                assert repr(root) == "<{} 'NoneType' (0 groups, 0 datasets, 0 metadata)>".format(writer.__name__)
        assert err.match('specify a file')


def test_file_path():
    # the URL is a string
    path = os.path.join(tempfile.gettempdir(), 'foo')
    if os.path.isfile(path):
        os.remove(path)

    for writer in writers:
        if writer is HDF5Writer and h5py is None:
            continue
        with writer(path) as root:
            assert root.file == path
            fill_root_with_data(root)
            assert_root_data(root)
            assert repr(root) == "<{} 'foo' (2 groups, 1 datasets, 2 metadata)>".format(writer.__name__)

        root2 = read(path)
        assert root2.file == path
        assert_root_data(root2)
        os.remove(path)


def test_exception_raised():
    # the file gets written even if an exception is raised
    path = os.path.join(tempfile.gettempdir(), 'bar')
    if os.path.isfile(path):
        os.remove(path)

    for writer in writers:
        if writer is HDF5Writer and h5py is None:
            continue
        with pytest.raises(ZeroDivisionError):
            with writer(path) as root:
                assert root.file == path
                assert repr(root) == "<{} 'bar' (0 groups, 0 datasets, 0 metadata)>".format(writer.__name__)
                divide = 1/0

        root2 = read(path)
        assert root2.file == path
        assert len(root2.metadata) == 0
        assert len(list(root2.groups())) == 0
        assert len(list(root2.datasets())) == 0
        os.remove(path)

    for writer in writers:
        if writer is HDF5Writer and h5py is None:
            continue
        with pytest.raises(ZeroDivisionError):
            with writer(path) as root:
                assert root.file == path
                fill_root_with_data(root)
                assert_root_data(root)
                assert repr(root) == "<{} 'bar' (2 groups, 1 datasets, 2 metadata)>".format(writer.__name__)
                divide = 1/0

        root2 = read(path)
        assert root2.file == path
        assert_root_data(root2)
        os.remove(path)


def test_update_context_kwargs():
    with StringIO() as buf:
        with JSONWriter() as root:
            root.add_metadata(one=1)
            root.create_dataset('dset', data=np.arange(9).reshape(3, 3))
            dtype_str = root.dset.dtype.str
            root.update_context_kwargs(file=buf, indent=None, separators=('|', ';'), sort_keys=True)
        file_info, value = buf.getvalue().splitlines()
        assert 'MSL JSONWriter' in file_info
        assert value == '{"dset";{"data";[[0|1|2]|[3|4|5]|[6|7|8]]|"dtype";"%s"}|"one";1}' % dtype_str



"""
A :class:`~msl.io.dataset.Dataset` that handles :mod:`logging` records.
"""
import logging
from datetime import datetime

import numpy as np

from .dataset import Dataset


class DatasetLogging(Dataset, logging.Handler):

    def __init__(self, name, parent, level=logging.NOTSET, attributes=None, logger=None, date_fmt=None, **kwargs):
        """A :class:`~msl.io.dataset.Dataset` that handles :mod:`logging` records.

        Do not instantiate directly. Create a new :class:`DatasetLogging` using
        :meth:`~msl.io.group.Group.create_dataset_logging`.

        Parameters
        ----------
        name : :class:`str`
            A name to associate with the :class:`~msl.io.dataset.Dataset`.
        parent : :class:`~msl.io.group.Group`
            The parent :class:`~msl.io.group.Group` to the :class:`~msl.io.dataset.Dataset`.
        level : :class:`int` or :class:`str`, optional
            The :ref:`logging level <levels>` to use.
        attributes : :class:`list` or :class:`tuple` of :class:`str`, optional
            The :ref:`attribute names <logrecord-attributes>` to include in the
            :class:`~msl.io.dataset.Dataset` for each :ref:`logging record <log-record>`.
        logger : :class:`~logging.Logger`, optional
            The :class:`~logging.Logger` that this :class:`DatasetLogging` object
            will be added to. If :data:`None` then it is added to the ``root`` :class:`~logging.Logger`.
        date_fmt : :class:`str`, optional
            The :class:`~datetime.datetime` :ref:`format code <strftime-strptime-behavior>`
            to use to represent the ``asctime`` :ref:`attribute <logrecord-attributes>` in.
        **kwargs
            Additional keyword arguments are passed to :class:`~msl.io.dataset.Dataset`.
            The default behaviour is to append every :ref:`logging record <log-record>`
            to the :class:`~msl.io.dataset.Dataset`. This guarantees that the size of the
            :class:`~msl.io.dataset.Dataset` is equal to the number of
            :ref:`logging records <log-record>` that were added to it. However, this behaviour
            can decrease the performance if many :ref:`logging records <log-record>` are
            added often because a copy of the data in the :class:`~msl.io.dataset.Dataset` is
            created for each :ref:`logging record <log-record>` that is added. You can improve
            the performance by specifying an initial size of the :class:`~msl.io.dataset.Dataset`
            by including a `shape` or a `size` keyword argument. This will also automatically
            create additional empty rows in the :class:`~msl.io.dataset.Dataset`, that is
            proportional to the size of the :class:`~msl.io.dataset.Dataset`, if the size of the
            :class:`~msl.io.dataset.Dataset` needs to be increased. If you do this then you will
            want to call :meth:`.remove_empty_rows` before writing :class:`DatasetLogging` to a
            file or interacting with the data in :class:`DatasetLogging` to remove the extra
            rows that were created.
        """
        if not attributes or not all(isinstance(a, str) for a in attributes):
            raise ValueError('Must specify attribute names as strings, got: {}'.format(attributes))

        self._logger = None
        self._attributes = tuple(attributes)
        self._uses_asctime = 'asctime' in attributes
        self._date_fmt = date_fmt

        if isinstance(level, str):
            level = getattr(logging, level)

        # these 3 keys in the metadata are used to distinguish a DatasetLogging
        # object from a regular Dataset object
        kwargs['logging_level'] = level
        kwargs['logging_level_name'] = logging.getLevelName(level)
        kwargs['logging_date_format'] = date_fmt
        self._dtype = np.dtype([(a, object) for a in attributes])

        self._auto_resize = 'size' in kwargs or 'shape' in kwargs
        if self._auto_resize:
            if 'size' in kwargs:
                kwargs['shape'] = (kwargs.pop('size'),)
            elif isinstance(kwargs['shape'], int):
                kwargs['shape'] = (kwargs['shape'],)

            shape = kwargs['shape']
            if len(shape) != 1:
                raise ValueError('Invalid shape {}, the number of dimensions must be 1'.format(shape))
            if shape[0] < 0:
                raise ValueError('Invalid shape {}'.format(shape))

        # call Dataset.__init__ before Handler.__init__ in case the Dataset cannot be created
        Dataset.__init__(self, name, parent, False, dtype=self._dtype, **kwargs)

        self._index = np.count_nonzero(self._data)
        if self._auto_resize and self._data.shape < kwargs['shape']:
            self._resize(new_allocated=kwargs['shape'][0])

        # the Handler will overwrite the self._name attribute, so we create a reference to the
        # `name` of the Dataset and then set the `name` of the Handler after it is initialized
        name = self._name
        logging.Handler.__init__(self, level=level)
        self.set_name(name)

        self.set_logger(logger or logging.getLogger())

    def __eq__(self, other):
        # Must override __eq__ because  calling root.addHandler(self) in
        # DatasetLogging.__init__ checks for the following condition:
        #   "if not (hdlr in self.handlers):"
        # and this could fail if adding multiple DatasetLogging instances
        # to the RootLogger and the Datasets are emtpy. Every DatasetLogging
        # instance must be unique because every Vertex is unique so this
        # method can simply return False.
        return False

    def __hash__(self):
        # need to override for linux and macOS running Python 3.7+
        return logging.Handler.__hash__(self)

    @property
    def attributes(self):
        """:class:`tuple` of :class:`str`: The :ref:`attribute names <logrecord-attributes>`
        used by the :class:`DatasetLogging` object.
        """
        return self._attributes

    @property
    def date_fmt(self):
        """:class:`str`: The :class:`~datetime.datetime` :ref:`format code <strftime-strptime-behavior>`
        that is used to represent the ``asctime`` :ref:`attribute <logrecord-attributes>` in.
        """
        return self._date_fmt

    @property
    def logger(self):
        """:class:`~logging.Logger`: The :class:`~logging.Logger` that this
        :class:`DatasetLogging` object is added to.
        """
        return self._logger

    def remove_empty_rows(self):
        """Remove empty rows from the :class:`~msl.io.dataset.Dataset`.

        If the :class:`DatasetLogging` object was initialized with a `shape` or a `size` keyword
        argument then the size of the :class:`~msl.io.dataset.Dataset` is always :math:`\\geq`
        to the number of :ref:`logging records <log-record>` that were added to it. Calling this
        method will remove the rows in the :class:`~msl.io.dataset.Dataset` that were not
        from a :ref:`logging record <log-record>`.
        """
        # don't use "is not None" since this does not work as expected
        self._data = self._data[self._data[self._dtype.names[0]] != None]

    def remove_handler(self):
        """Remove this class's :class:`~logging.Handler` from the associated :class:`~logging.Logger`.

        After calling this method :ref:`logging records <log-record>` are no longer
        added to the :class:`~msl.io.dataset.Dataset`.
        """
        if self._logger is not None:
            self._logger.removeHandler(self)

    def set_logger(self, logger):
        """Add this class's :class:`~logging.Handler` to a :class:`~logging.Logger`.

        Parameters
        ----------
        logger : :class:`~logging.Logger`
            The :class:`~logging.Logger` to add this class's :class:`~logging.Handler` to.
        """
        if not isinstance(logger, logging.Logger):
            raise TypeError('Must be a logging.Logger object')

        level = self.metadata.logging_level
        if logger.level == 0 or logger.level > level:
            logger.setLevel(level)

        self.remove_handler()
        logger.addHandler(self)
        self._logger = logger

    def emit(self, record):
        """Overrides the :meth:`~logging.Handler.emit` method."""
        record.message = record.getMessage()
        if self._uses_asctime:
            record.asctime = datetime.fromtimestamp(record.created).strftime(self._date_fmt)
        latest = tuple(record.__dict__[a] for a in self._attributes)
        row = np.asarray(latest, dtype=self._dtype)
        if self._auto_resize:
            if self._index >= self._data.size:
                self._resize()
            self._data[self._index] = row
            self._index += 1
        else:
            self._data = np.append(self._data, row)

    def _resize(self, new_allocated=None):
        # Over-allocates proportional to the size of the ndarray, making room
        # for additional growth. This follows the over-allocating procedure that
        # Python uses when appending to a list object, see `list_resize` in
        # https://github.com/python/cpython/blob/master/Objects/listobject.c
        if new_allocated is None:
            new_size = self._data.size + 1
            new_allocated = new_size + (new_size >> 3) + (3 if new_size < 9 else 6)

        # don't use self._data.resize() because that fills the newly-created rows
        # with 0 and want to fill the new rows with None to be explicit that the
        # new rows are not associated with logging records
        array = np.empty((new_allocated,), dtype=self._dtype)
        array[:self._data.size] = self._data
        self._data = array



"""
Reader for the Detector Responsivity System from Light Standards at MSL.
"""
import re
from datetime import datetime

import numpy as np

from ..base import Reader
from ..utils import register


@register
class DRSReader(Reader):
    """Reader for the Detector Responsivity System from Light Standards at MSL."""

    @staticmethod
    def can_read(file, **kwargs):
        """Checks if the first line starts with ``DRS`` and ends with ``Shindo``."""
        if Reader.get_extension(file).lower() != '.dat':
            return False
        line = Reader.get_lines(file, 1)[0]
        if line.startswith('DRS') and line.endswith('Shindo'):
            return True
        if line.startswith('"DRS') and line.endswith('Shindo"'):
            return True
        return False

    def read(self, **kwargs):
        """Reads the *.DAT* and corresponding *.LOG* file.

        Parameters
        ----------
        **kwargs
            All key-value pairs are ignored.
        """
        self._default_alias = {'l(nm)': 'wavelength'}

        self._lines_dat = self.get_lines(self.file, remove_empty_lines=True)
        self._num_lines_dat = len(self._lines_dat)

        self._lines_log = self.get_lines(self.file[:-3]+'LOG', remove_empty_lines=True)
        self._num_lines_log = len(self._lines_log)

        num_runs = 0
        self._index_dat = 0
        self._index_log = 0
        while self._index_dat < self._num_lines_dat:
            if 'DRS' in self._lines_dat[self._index_dat]:
                num_runs += 1
                group = self.create_group('run%d' % num_runs)
                self._read_run(group)
            else:
                self._index_dat += 1
                self._index_log += 1

    def _read_run(self, group):
        group.add_metadata(**self._get_run_metadata())

        # include all Scans
        scan_number = 0
        while self._index_dat < self._num_lines_dat and self._lines_dat[self._index_dat].startswith('Scan'):
            scan_number += 1
            sub = group.create_group('scan%d' % scan_number)
            self._get_scan_dat(sub)
            self._get_scan_log(sub)
            self._index_dat += 1
            self._index_log += 1

    def _get_run_metadata(self):
        # add the metadata from the LOG file

        assert 'DRS' in self._lines_log[self._index_log], self._lines_log[self._index_log]

        meta = dict()
        meta['version'] = re.search(r'version (\w+\s+\d{4})', self._lines_log[self._index_log]).group(1)
        meta['author'] = re.search(r'by\s(.*)', self._lines_log[self._index_log]).group(1)

        # skip lines until we reach the info about the wavelength scan values
        while not self._lines_log[self._index_log].startswith('Wavelength start'):
            self._index_log += 1

        # the info about the wavelength range
        assert self._lines_log[self._index_log].startswith('Wavelength start(nm)'), self._lines_log[self._index_log]
        meta['wavelength_units'] = 'nm'
        meta['wavelength_start'] = float(self._lines_log[self._index_log].split(':')[1])
        self._index_log += 1
        meta['wavelength_end'] = float(self._lines_log[self._index_log].split(':')[1])
        self._index_log += 1
        meta['wavelength_increment'] = float(self._lines_log[self._index_log].split(':')[1])
        self._index_log += 1
        if self._lines_log[self._index_log].startswith('Extra'):
            line_split = self._lines_log[self._index_log][len('Extra wavelength (nm):'):].split(';')
            extras = [val for val in line_split if val.strip()]
            meta['extra_wavelengths_nm'] = list(map(float, extras))
            self._index_log += 1

        # the info about the grating used and the bandwidth
        grating = re.search(r'at: (\d+) nm', self._lines_log[self._index_log])
        bw = re.search(r'correction: ([\d.]+) nm', self._lines_log[self._index_log])
        if grating and bw:
            meta['grating_nm'] = int(grating.group(1))
            meta['bandwidth_nm'] = float(bw.group(1))
        else:
            self._index_log -= 1  # go back a line

        # get the number of devices
        self._index_log += 1
        assert self._lines_log[self._index_log].startswith('Numbers of'), self._lines_log[self._index_log]
        # the info about the devices under test
        self._index_log += 1
        meta['devices'] = dict()
        self._names = [dev.strip().upper() for dev in self._lines_log[self._index_log].split('\t') if dev]
        if self._names[0].endswith('I'):
            skip = 1
            self._names = self._names[1:]
        else:
            skip = 0
        for name in self._names:
            meta['devices'][name] = dict()
        for label in ['Name', 'Tprobe', 'Channel']:
            self._index_log += 1
            line_split = [item.strip() for item in self._lines_log[self._index_log].split('\t') if item]
            device_key = label if skip == 0 else line_split[0]
            for index, name in enumerate(self._names):
                meta['devices'][name][device_key] = line_split[index + skip]

        # some LOG files have info about the lab temperature
        self._index_log += 1
        if self._lines_log[self._index_log].startswith('Laboratory Temperature'):
            line_split = [item.strip() for item in self._lines_log[self._index_log].split(';')]
            meta['Laboratory Temperature {}'.format(line_split[1])] = float(line_split[2])
            self._index_log += 1

        # some have info about the
        while 'R0=' in self._lines_log[self._index_log]:
            prt = self._lines_log[self._index_log].split('\t')
            prt_dict = {
                'R0': float(prt[1][3:]),
                'A': float(prt[3][2:]),
                'B': float(prt[5][2:]),
                'value': float(prt[7]),
            }
            meta[prt[0][:-1].strip()] = prt_dict
            self._index_log += 1

        if self._lines_log[self._index_log].startswith('Integration time'):
            meta['Integration time (s)'] = float(self._lines_log[self._index_log].split(':')[1])
            self._index_log += 1

        assert self._lines_log[self._index_log].startswith('Source Monitoring'), self._lines_log[self._index_log]
        meta['Source Monitoring'] = self._lines_log[self._index_log][len('Source Monitoring'):].strip()
        self._index_log += 1
        while 'DVM RANGE' in self._lines_log[self._index_log]:
            dvm = re.search(r'(\w+)::DVM RANGE\(V\)=\s+(.*\d+)\s+;\s+DVM NPLC=\s+(\d+)', self._lines_log[self._index_log])
            meta['{}_dvm_range'.format(dvm.group(1))] = float(dvm.group(2))
            meta['{}_dvm_nplc'.format(dvm.group(1))] = float(dvm.group(3))
            self._index_log += 1

        assert self._lines_log[self._index_log].startswith('Stray light filters'), self._lines_log[self._index_log]
        meta['Stray light filters'] = self._lines_log[self._index_log][len('Stray light filters'):].strip()

        # get the comment
        self._index_log += 1
        assert self._lines_log[self._index_log] == 'Comment:', self._lines_log[self._index_log]
        self._index_log += 1
        comment = []
        while not self._lines_log[self._index_log].startswith('Scan'):
            comment.append(self._lines_log[self._index_log])
            self._index_log += 1
        meta['comment'] = '\n'.join(comment)

        # move the DAT index to the start of the Scan
        while not self._lines_dat[self._index_dat].startswith('Scan'):
            self._index_dat += 1

        return meta

    def _get_scan_dat(self, group):
        meta = dict()

        assert self._lines_dat[self._index_dat].startswith('Scan'), self._lines_dat[self._index_dat]
        meta['start_time'] = self._to_datetime(self._lines_dat[self._index_dat])

        # get the field names to use for the numpy array
        aliases = self._default_alias.copy()
        for name in self._names:
            aliases[name] = group.parent.metadata.devices[name]['Name']

        self._index_dat += 1
        header = [item.strip() for item in self._lines_dat[self._index_dat].split('\t') if item]
        fieldnames = []
        for h in header:
            name = h
            for key, value in aliases.items():
                if key in h:
                    name = h.replace(key, value)
                    if name in fieldnames:
                        name += '-({})'.format(key[-1])
                    break
            fieldnames.append(name)

        # get the DAT data
        self._index_dat += 1
        data = []
        while self._index_dat < self._num_lines_dat and not self._lines_dat[self._index_dat].startswith('End'):
            data.append(tuple([float(item) for item in self._lines_dat[self._index_dat].split('\t') if item]))
            self._index_dat += 1

        if self._index_dat < self._num_lines_dat:  # otherwise the scan was aborted early
            assert self._lines_dat[self._index_dat].startswith('End of scan'), self._lines_dat[self._index_dat]
            meta['end_time'] = self._to_datetime(self._lines_dat[self._index_dat])

        # sometimes there was more data columns than header columns
        if data:
            i = 1
            while len(data[0]) > len(fieldnames):
                fieldnames.append('UNKNOWN-%d' % i)
                i += 1

        group.create_dataset(
            'dat',
            data=data,
            dtype=[(name, float) for name in fieldnames],
            **meta
        )

    def _get_scan_log(self, group):
        meta = dict()

        assert self._lines_log[self._index_log].startswith('Scan'), self._lines_log[self._index_log]
        meta['start_time'] = self._to_datetime(self._lines_log[self._index_log])

        # get the field names to use for the numpy array
        aliases = self._default_alias.copy()
        for name in self._names:
            aliases[name] = group.parent.metadata.devices[name]['Name']

        self._index_log += 1
        header = [item.strip() for item in self._lines_log[self._index_log].split('\t') if item.strip()]
        assert header[0] == 'l(nm)', self._lines_log[self._index_log]
        fieldnames = []
        for h in header:
            if h == 'Name':
                continue
            name = h
            for key, value in aliases.items():
                if key in h:
                    name = h.replace(key, value)
                    break
            if name not in fieldnames:
                fieldnames.append(name)

        # get the LOG data
        self._index_log += 1
        n = len(header)
        data = [[] for _ in self._names]
        while self._index_log < self._num_lines_log and not self._lines_log[self._index_log].startswith('End'):
            for i, name in enumerate(self._names):
                values = []
                items = [item for item in self._lines_log[self._index_log].split('\t') if item]
                for j, item in enumerate(items):
                    if j == 0:
                        values.append(float(item[:-2]))
                    elif j == 1:
                        assert item == aliases[name], 'expect Name={}, got {}'.format(aliases[name], item)
                    elif j >= n:
                        break
                    else:
                        values.append(float(item))

                # fill in the blank columns with NaN
                while len(values) < len(fieldnames):
                    values.append(np.NaN)

                data[i].append(tuple(values))
                self._index_log += 1

        if self._index_log < self._num_lines_log:  # otherwise the scan was aborted early
            assert self._lines_log[self._index_log].startswith('End of scan'), self._lines_log[self._index_log]
            meta['end_time'] = self._to_datetime(self._lines_log[self._index_log])

        for i, name in enumerate(self._names):
            vertex_name = 'log-' + aliases[name]
            vertex_name = vertex_name.replace('.', '')  # cannot contain a "."
            if vertex_name in group:
                vertex_name += '-({})'.format(name[-1])

            group.create_dataset(
                vertex_name,
                data=data[i],
                dtype=[(name, float) for name in fieldnames],
                **meta
            )

    def _to_datetime(self, line):
        string = line.split(',')[1].replace('p.m.', 'PM').replace('a.m.', 'AM').strip()
        try:
            return datetime.strptime(string, '%I:%M %p %d/%m/%Y')
        except ValueError:
            return datetime.strptime(string, '%d/%m/%Y %I:%M %p')



"""
Base classes for all :class:`Reader`\\s. and :class:`Writer`\\s.
"""
import codecs
import itertools
import os

from .constants import IS_PYTHON2
from .group import Group
from .utils import get_basename


class Root(Group):

    def __init__(self, file, **metadata):
        """The root_ vertex in a tree_.

        .. _root: https://en.wikipedia.org/wiki/Tree_(graph_theory)#Rooted_tree
        .. _tree: https://en.wikipedia.org/wiki/Tree_(graph_theory)

        Parameters
        ----------
        file : :term:`path-like <path-like object>` or :term:`file-like <file object>`
            The file object to associate with the :class:`~msl.io.base.Root`.
        **metadata
            Key-value pairs that can be used as :class:`~msl.io.metadata.Metadata`
            for the :class:`~msl.io.base.Root`.
        """
        super(Group, self).__init__('/', None, False, **metadata)
        self._file = file

    def __repr__(self):
        b = get_basename(self._file)
        g = len(list(self.groups()))
        d = len(list(self.datasets()))
        m = len(self.metadata)
        return '<{} {!r} ({} groups, {} datasets, {} metadata)>'.\
            format(self.__class__.__name__, b, g, d, m)

    @property
    def file(self):
        """:term:`path-like <path-like object>` or :term:`file-like <file object>`: The file
        object that is associated with the :class:`Root`."""
        return self._file

    def tree(self, indent=2):
        """A representation of the `tree structure`_ of all :class:`~msl.io.group.Group`\\s
        and :class:`~msl.io.dataset.Dataset`\\s that are in :class:`Root`.

        .. _tree structure: https://en.wikipedia.org/wiki/Tree_structure

        Parameters
        ----------
        indent : :class:`int`, optional
            The amount of indentation to add for each recursive level.

        Returns
        -------
        :class:`str`
            The `tree structure`_.
        """
        return repr(self) + '\n' + \
            '\n'.join(' ' * (indent * k.count('/')) + repr(v) for k, v in sorted(self.items()))


class Writer(Root):

    def __init__(self, file=None, **metadata):
        """
        Parameters
        ----------
        file : :term:`path-like <path-like object>` or :term:`file-like <file object>`, optional
            The file to write the data to. Can also be specified in the :meth:`write` method.
        **metadata
            Key-value pairs that are used as :class:`~msl.io.metadata.Metadata`
            of the :class:`~msl.io.base.Root`.
        """
        super(Writer, self).__init__(file, **metadata)
        self._context_kwargs = {}

    def set_root(self, root):
        """Set a new :class:`Root` for the :class:`Writer`.

        .. attention::
           This will clear the :class:`~msl.io.metadata.Metadata` of the :class:`Writer`
           and all :class:`~msl.io.group.Group`\\s and :class:`~msl.io.dataset.Dataset`\\s
           that the :class:`Writer` currently contains. The `file` that was specified when
           the :class:`Writer` was created does not change.

        Parameters
        ----------
        root : :class:`Root`
            The new :class:`Root` for the :class:`Writer`.
        """
        if not isinstance(root, Group):  # it is okay to pass in any Group object
            raise TypeError('Must pass in a Root object, got {!r}'.format(root))
        self.clear()
        self.metadata.clear()
        self.add_metadata(**root.metadata)
        if root:  # only do this if there are Groups and/or Datasets in the new root
            self.add_group('', root)

    def update_context_kwargs(self, **kwargs):
        """
        When a :class:`Writer` is used as a :ref:`context manager <with>` the
        :meth:`write` method is automatically called when exiting the
        :ref:`context manager <with>`. You can specify the keyword arguments
        that will be passed to the :meth:`write` method by calling
        :meth:`update_context_kwargs` with the appropriate key-value pairs
        before the :ref:`context manager <with>` exits. You can call this
        method multiple times since the key-value pairs get added to the
        underlying :class:`dict` (via :meth:`dict.update`) that contains
        all keyword arguments which are passed to the :meth:`write` method.
        """
        self._context_kwargs.update(**kwargs)

    def write(self, file=None, root=None, **kwargs):
        """Write to a file.

        .. important::
           You must override this method.

        Parameters
        ----------
        file : :term:`path-like <path-like object>` or :term:`file-like <file object>`, optional
            The file to write the `root` to. If :data:`None` then uses the
            `file` value that was specified when the :class:`~msl.io.base.Writer`
            was instantiated.
        root : :class:`~msl.io.base.Root`, optional
            Write the `root` object in the file format of this :class:`~msl.io.base.Writer`.
            This is useful when converting between different file formats.
        **kwargs
            Additional key-value pairs to use when writing the file.
        """
        raise NotImplementedError

    def save(self, file=None, root=None, **kwargs):
        """Alias for :meth:`write`."""
        self.write(file=file, root=root, **kwargs)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        # always write the Root to the file even if
        # an exception was raised in the context manager
        self.write(**self._context_kwargs)


class Reader(Root):

    def __init__(self, file):
        """
        Parameters
        ----------
        file : :term:`path-like <path-like object>` or :term:`file-like <file object>`
            The file to read.
        """
        super(Reader, self).__init__(file)

    @staticmethod
    def can_read(file, **kwargs):
        """Whether this :class:`~msl.io.base.Reader` can read the file specified by `file`.

        .. important::
           You must override this method.

        Parameters
        ----------
        file : :term:`path-like <path-like object>` or :term:`file-like <file object>`
            The file to check whether the :class:`~msl.io.base.Reader` can read it.
        **kwargs
            Key-value pairs that the :class:`~msl.io.base.Reader` class may need
            when checking if it can read the `file`.

        Returns
        -------
        :class:`bool`
            Either :data:`True` (can read) or :data:`False` (cannot read).
        """
        return False

    def read(self, **kwargs):
        """Read the file.

        The file can be accessed by the :attr:`~msl.io.base.Root.file`
        property of the :class:`~msl.io.base.Reader`, i.e., ``self.file``

        .. important::
           You must override this method.

        Parameters
        ----------
        **kwargs
            Key-value pairs that the :class:`~msl.io.base.Reader` class may need
            when reading the file.
        """
        raise NotImplementedError

    @staticmethod
    def get_lines(file, *args, **kwargs):
        """Return lines from a file.

        Parameters
        ----------
        file : :term:`path-like <path-like object>` or :term:`file-like <file object>`
            The file to read lines from.
        *args : :class:`int` or :class:`tuple` of :class:`int`, optional
            The line(s) in the file to get.

            Examples:

            * ``get_lines(file)`` :math:`\\rightarrow` returns all lines

            * ``get_lines(file, 5)`` :math:`\\rightarrow` returns the first 5 lines

            * ``get_lines(file, -5)`` :math:`\\rightarrow` returns the last 5 lines

            * ``get_lines(file, 2, 4)`` :math:`\\rightarrow` returns lines 2, 3 and 4

            * ``get_lines(file, 4, -1)`` :math:`\\rightarrow` skips the first 3 lines
              and returns the rest

            * ``get_lines(file, 2, -2)`` :math:`\\rightarrow` skips the first and last
              lines and returns the rest

            * ``get_lines(file, -4, -2)`` :math:`\\rightarrow` returns the fourth-,
              third- and second-last lines

            * ``get_lines(file, 1, -1, 6)`` :math:`\\rightarrow` returns every sixth
              line in the file

        **kwargs

            * ``remove_empty_lines`` : :class:`bool`

              Whether to remove all empty lines. Default is :data:`False`.

            * ``encoding`` : :class:`str`

              The name of the encoding to use to decode the file. The default is ``'utf-8'``.

            * ``errors`` : :class:`str`

              An optional string that specifies how encoding errors are to be handled.
              Either ``'strict'`` or ``'ignore'``. The default is ``'strict'``.

        Returns
        -------
        :class:`list` of :class:`str`
            The lines from the file. Trailing whitespace is stripped from each line.
        """
        remove_empty_lines = kwargs.pop('remove_empty_lines', False)

        if 'encoding' not in kwargs:
            kwargs['encoding'] = 'utf-8'

        if IS_PYTHON2:
            opener = codecs.open  # this allows the encoding and errors kwargs to be used
        else:
            opener = open

        # want the "stop" line to be included
        if (len(args) > 1) and (args[1] is not None) and (args[1] < 0):
            if args[1] == -1:
                args = (args[0], None) + args[2:]
            else:
                args = (args[0],) + (args[1] + 1,) + args[2:]

        # want the "start" line to be included
        if (len(args) > 1) and (args[0] is not None) and (args[0] > 0):
            args = (args[0] - 1,) + args[1:]

        is_file_like = hasattr(file, 'tell')

        # itertools.islice does not support negative indices, but want to allow
        # getting the last "N" lines from a file.
        if any(val < 0 for val in args if val):
            def get(_file):
                return [line.rstrip() for line in _file.readlines()]

            if is_file_like:
                position = file.tell()
                lines = get(file)
                file.seek(position)
            else:
                with opener(file, 'r', **kwargs) as f:
                    lines = get(f)

            if len(args) == 1:
                lines = lines[args[0]:]
            elif len(args) == 2:
                lines = lines[args[0]:args[1]]
            else:
                lines = lines[args[0]:args[1]:args[2]]

        else:
            if not args:
                args = (None,)

            def get(_file):
                return [line.rstrip() for line in itertools.islice(_file, *args)]

            if is_file_like:
                position = file.tell()
                lines = get(file)
                file.seek(position)
            else:
                with opener(file, 'r', **kwargs) as f:
                    lines = get(f)

        if remove_empty_lines:
            return [line for line in lines if line]

        return lines

    @staticmethod
    def get_bytes(file, *args):
        """Return bytes from a file.

        Parameters
        ----------
        file : :term:`path-like <path-like object>` or :term:`file-like <file object>`
            The file to read bytes from.
        *args : :class:`int` or :class:`tuple` of :class:`int`, optional
            The position(s) in the file to retrieve bytes from.

            Examples:

            * ``get_bytes(file)`` :math:`\\rightarrow` returns all bytes

            * ``get_bytes(file, 5)`` :math:`\\rightarrow` returns the first 5 bytes

            * ``get_bytes(file, -5)`` :math:`\\rightarrow` returns the last 5 bytes

            * ``get_bytes(file, 5, 10)`` :math:`\\rightarrow` returns bytes 5 through
              10 (inclusive)

            * ``get_bytes(file, 3, -1)`` :math:`\\rightarrow` skips the first 2 bytes
              and returns the rest

            * ``get_bytes(file, -8, -4)`` :math:`\\rightarrow` returns the eighth-
              through fourth-last bytes (inclusive)

            * ``get_bytes(file, 1, -1, 2)`` :math:`\\rightarrow` returns every other byte

        Returns
        -------
        :class:`bytes`
            The bytes from the file.
        """
        is_file_like = hasattr(file, 'tell')
        if is_file_like:
            position = file.tell()
            file.seek(0, os.SEEK_END)
            size = file.tell()
            file.seek(position)
        else:
            try:
                size = os.path.getsize(file)
            except OSError:
                # A file on a mapped network drive can raise the following:
                #   [WinError 87] The parameter is incorrect
                # for Python 3.5, 3.6 and 3.7. Also, calling os.path.getsize
                # on a file on a mapped network drive could return 0
                # (without raising OSError) on Python 3.8 and 3.9, which is
                # why we set size=0 on an OSError
                size = 0

            if size == 0:
                with open(file, mode='rt') as fp:
                    fp.seek(0, os.SEEK_END)
                    size = fp.tell()

        if not args:
            start, stop, step = 0, size, 1
        elif len(args) == 1:
            start, stop, step = 0, args[0], 1
            if stop is None:
                stop = size
            elif stop < 0:
                start, stop = size + stop + 1, size
        elif len(args) == 2:
            start, stop, step = args[0] or 0, args[1], 1
            if stop is None or stop == -1:
                stop = size
        else:
            start, stop, step = args[0] or 0, args[1] or size, args[2] or 1

        if start < 0:
            start = max(size + start, 0)
        elif start > 0:
            start -= 1
        start = min(size, start)

        if stop < 0:
            stop += size + 1
        stop = min(size, stop)

        if is_file_like:
            position = file.tell()
            _file = file
        else:
            _file = open(file, mode='rb')

        _file.seek(start)
        data = _file.read(max(0, stop - start))
        if step != 1:
            data = data[::step]

        if is_file_like:
            file.seek(position)
        else:
            _file.close()

        return data

    @staticmethod
    def get_extension(file):
        """Return the extension of the file.

        Parameters
        ----------
        file : :term:`path-like <path-like object>` or :term:`file-like <file object>`
            The file to get the extension of.

        Returns
        -------
        :class:`str`
            The extension, including the ``'.'``.
        """
        try:
            return os.path.splitext(file)[1]
        except (TypeError, AttributeError):
            try:
                return Reader.get_extension(file.name)
            except AttributeError:
                return ''



"""
Provides information about other data.
"""
try:
    # this try..except block fixes:
    #   DeprecationWarning: Using or importing the ABCs from 'collections' instead
    #   of from 'collections.abc' is deprecated, and in 3.8 it will stop working
    from collections.abc import MutableMapping
except ImportError:
    from collections import MutableMapping

import numpy as np

from .dictionary import Dictionary


class Metadata(Dictionary):

    def __init__(self, read_only, vertex_name, **kwargs):
        """Provides information about other data.

        Do not instantiate directly. A :class:`Metadata` object is created automatically
        when :meth:`~msl.io.group.Group.create_dataset` or :meth:`~msl.io.group.Group.create_group`
        is called.

        Parameters
        ----------
        read_only : :class:`bool`
            Whether :class:`Metadata` is to be accessed in read-only mode.
        vertex_name : :class:`str`
            The name of the :class:`~msl.io.vertex.Vertex` that :class:`Metadata` is associated with.
        **kwargs
            Key-value pairs that will be used to create the :class:`.Dictionary`.
        """
        super(Metadata, self).__init__(read_only, **kwargs)
        self._vertex_name = vertex_name

    def __repr__(self):
        return '<Metadata {!r} {}>'.format(self._vertex_name, super(Metadata, self).__repr__())

    def __getitem__(self, item):
        try:
            value = self._mapping[item]
        except KeyError:
            pass  # raise a more detailed error message below
        else:
            if isinstance(value, MutableMapping):
                return Metadata(self._read_only, self._vertex_name, **value)
            return value
        self._raise_key_error(item)

    def __delattr__(self, item):
        self._raise_if_read_only()
        try:
            del self._mapping[item]
            return
        except KeyError as e:
            msg = str(e)
        raise AttributeError(msg)

    def __getattr__(self, item):
        try:
            return self.__getitem__(item)
        except KeyError as e:
            msg = str(e)
        raise AttributeError(msg)

    def __setattr__(self, item, value):
        if item.endswith('read_only'):
            val = bool(value)
            self.__dict__['_read_only'] = val
            try:
                # make all numpy ndarray's read only also
                for obj in self.__dict__['_mapping'].values():
                    if isinstance(obj, np.ndarray):
                        obj.setflags(write=not val)
            except KeyError:
                pass
        elif item == '_mapping' or item == '_vertex_name':
            self.__dict__[item] = value
        else:
            self._raise_if_read_only()
            self._mapping[item] = value

    def copy(self, read_only=None):
        """Create a copy of the :class:`Metadata`.

        Parameters
        ----------
        read_only : :class:`bool`, optional
            Whether the copy should be created in read-only mode. If :data:`None` then
            creates a copy using the mode for the :class:`Metadata` that is being copied.

        Returns
        -------
        :class:`Metadata`
            A copy of the :class:`Metadata`.
        """
        return Metadata(self._mode(read_only), self._vertex_name, **self._mapping)

    def fromkeys(self, seq, value=None, read_only=None):
        """Create a new :class:`Metadata` object with keys from `seq` and values set to `value`.

        Parameters
        ----------
        seq
            Any iterable object that contains the names of the keys.
        value : :class:`object`, optional
            The default value to use for each key.
        read_only : :class:`bool`, optional
            Whether the returned object should be created in read-only mode. If
            :data:`None` then uses the mode for the :class:`Metadata` that is used
            to call this method.

        Returns
        -------
        :class:`Metadata`
            A new :class:`Metadata` object.
        """
        return Metadata(self._mode(read_only), self._vertex_name, **dict((key, value) for key in seq))

    def _mode(self, read_only):
        return self._read_only if read_only is None else read_only



import sys

import numpy as np
import pytest

from msl.io.dataset import Dataset


def test_instantiate():
    dset = Dataset(name='/data', parent=None, read_only=True, shape=(10, 10))
    assert dset.name == '/data'
    assert len(dset) == 10
    assert dset.size == 100
    assert dset.dtype == float
    assert dset.dtype.names is None

    dset = Dataset(name='dataset 1', parent=None, read_only=True, shape=(100,), dtype=int)
    assert dset.name == 'dataset 1'
    assert len(dset) == 100
    assert dset.size == 100
    assert dset.dtype == int
    assert dset.dtype.names is None

    dset = Dataset(name='mixed', parent=None, read_only=True, shape=(100,), dtype=[('x', float), ('y', int), ('z', str)])
    assert dset.name == 'mixed'
    assert len(dset) == 100
    assert dset.size == 100
    assert len(dset['x']) == 100
    assert dset.dtype[0] == float
    assert len(dset['y']) == 100
    assert dset.dtype[1] == int
    assert len(dset['z']) == 100
    assert dset.dtype[2] == str
    assert dset.dtype.names == ('x', 'y', 'z')

    dset = Dataset(name='xxx', parent=None, read_only=True, data=[1, 2, 3])
    assert len(dset) == 3
    assert dset[0] == 1
    assert dset[1] == 2
    assert dset[2] == 3
    d = dset[:]
    assert len(d) == 3
    assert d[0] == 1
    assert d[1] == 2
    assert d[2] == 3
    d = dset[::2]
    assert len(d) == 2
    assert d[0] == 1
    assert d[1] == 3


def test_metadata():
    dset = Dataset(name='d', parent=None, read_only=False, shape=(100,),
                   dtype=int, order='F', temperature=21.3, lab='msl', x=-1)

    # 'name' is absorbed by Vertex
    # 'shape', 'dtype' and 'order' are kwargs that are absorbed by numpy

    assert len(dset.metadata) == 3
    assert dset.metadata['temperature'] == 21.3
    assert dset.metadata['lab'] == 'msl'
    assert dset.metadata['x'] == -1

    assert not dset.metadata.read_only

    dset.add_metadata(one=1, two=2, three=3)
    assert len(dset.metadata) == 6
    assert dset.metadata['one'] == 1
    assert dset.metadata['two'] == 2
    assert dset.metadata['three'] == 3


def test_field_access_as_attribute():
    # no names defined in the dtype
    dset = Dataset(name='data', parent=None, read_only=False, shape=(3, 3))
    assert len(dset) == 3
    assert dset.shape == (3, 3)
    assert dset.dtype == float
    assert dset.dtype.names is None

    with pytest.raises(AttributeError):
        _ = dset.there_are_no_field_names

    # names are defined in the dtype
    dset = Dataset(name='data', parent=None, read_only=False, shape=(100,),
                   dtype=[('x', float), ('y', int), ('z', str)])
    assert len(dset['x']) == 100
    assert len(dset.x) == 100
    assert len(dset['y']) == 100
    assert len(dset.y) == 100
    dset.y[:] = 1
    for val in dset.y:
        assert val == 1
    dset.x = np.arange(100, 200)
    assert np.array_equal(dset.x + dset.y, np.arange(101, 201))
    assert len(dset['z']) == 100
    assert len(dset.z) == 100
    assert dset['z'][0] == ''
    assert dset.z[0] == ''
    assert dset.dtype.names == ('x', 'y', 'z')


def test_read_only():
    dset = Dataset(name='my data', parent=None, read_only=True, shape=(100,), dtype=int)
    assert dset.name == 'my data'
    assert len(dset) == 100
    assert dset.read_only
    assert dset.metadata.read_only

    # cannot modify data
    with pytest.raises(ValueError):
        dset[:] = 1

    # cannot modify data
    with pytest.raises(ValueError):
        dset[0] = 1

    # make writable
    dset.read_only = False
    assert not dset.read_only
    assert not dset.metadata.read_only

    # can modify data
    dset[:] = 1
    assert dset[0] == 1

    # make read only again
    dset.read_only = True
    assert dset.read_only
    assert dset.metadata.read_only

    # cannot modify data
    with pytest.raises(ValueError):
        dset[:] = 1

    # can make a dataset writeable but the metadata read-only
    dset.read_only = False
    assert not dset.read_only
    assert not dset.metadata.read_only
    dset.metadata.read_only = True
    assert not dset.read_only
    assert dset.metadata.read_only
    dset[:] = 1
    with pytest.raises(ValueError):
        dset.add_metadata(some_more_info=1)


def test_copy():
    orig = Dataset(name='abcdefg', parent=None, read_only=True, shape=(10,), dtype=int, voltage=1.2, current=5.3)

    assert orig.read_only
    assert orig.name == 'abcdefg'

    copy = orig.copy()
    assert isinstance(copy, Dataset)
    assert copy.read_only
    assert copy.metadata.read_only
    assert copy.name == 'abcdefg'
    for i in range(10):
        assert orig[i] == copy[i]
    assert orig.metadata['voltage'] == copy.metadata['voltage']
    assert orig.metadata['current'] == copy.metadata['current']

    copy.read_only = False

    assert not copy.read_only
    assert not copy.metadata.read_only
    assert orig.read_only
    assert orig.metadata.read_only

    val = 7 if orig[1] != 7 else 8
    copy[1] = val
    assert copy[1] == val
    assert orig[1] != copy[1]


def test_string_representation():
    dset = Dataset(name='abcd', parent=None, data=[[1, 2], [3, 4]], read_only=True, foo='bar')

    assert repr(dset) in ["<Dataset 'abcd' shape=(2, 2) dtype='<f8' (1 metadata)>",
                          "<Dataset 'abcd' shape=(2L, 2L) dtype='<f8' (1 metadata)>"]

    assert str(dset) == ('array([[1., 2.],\n'
                         '       [3., 4.]])')

    assert str(dset.metadata) == "<Metadata 'abcd' {'foo': 'bar'}>"

    # just for fun, test more index access
    assert dset[0, 0] + dset[0, 1] == 3
    assert all(dset[:, 0] + dset[:, 1] == [3, 7])
    assert all(dset[0, :] + dset[1, :] == [4, 6])


def test_ndarray_attribute():
    dset = Dataset(name='abcd', parent=None, data=[[1, 2], [3, 4]], read_only=True)

    as_list = dset.tolist()
    assert isinstance(as_list, list)
    assert dset.tolist() == [[1, 2], [3, 4]]
    assert dset.flatten().tolist() == [1, 2, 3, 4]

    assert dset.max() == 4
    assert dset.min() == 1
    assert all(dset.max(axis=1) == [2, 4])
    assert all(dset.max(axis=0) == [3, 4])


def test_scalar():
    dset = Dataset(name='abcd', parent=None, data=5, read_only=True)
    assert len(dset) == 1
    assert dset.shape == ()
    assert dset.size == 1
    assert dset.data == 5.0


def test_add():
    d1 = Dataset(name='/d1', parent=None, read_only=True, data=[1, 2, 3])
    d2 = Dataset(name='/d2', parent=None, read_only=True, data=[4, 5, 6])

    for rhs in ([4, 5, 6], d2):
        result = d1 + rhs
        assert isinstance(result, np.ndarray)
        assert np.array_equal(result, np.array([5., 7., 9.]))

    for lhs in ([4, 5, 6], d2):
        result = lhs + d1
        assert isinstance(result, np.ndarray)
        assert np.array_equal(result, np.array([5., 7., 9.]))


def test_sub():
    d1 = Dataset(name='/d1', parent=None, read_only=True, data=[1, 2, 3])
    d2 = Dataset(name='/d2', parent=None, read_only=True, data=[4, 5, 6])

    for rhs in ([4, 5, 6], d2):
        result = d1 - rhs
        assert isinstance(result, np.ndarray)
        assert np.array_equal(result, np.array([-3., -3., -3.]))

    for lhs in ([4, 5, 6], d2):
        result = lhs - d1
        assert isinstance(result, np.ndarray)
        assert np.array_equal(result, np.array([3., 3., 3.]))


def test_mul():
    d1 = Dataset(name='/d1', parent=None, read_only=True, data=[1, 2, 3])
    d2 = Dataset(name='/d2', parent=None, read_only=True, data=[4, 5, 6])

    for rhs in ([4, 5, 6], d2):
        result = d1 * rhs
        assert isinstance(result, np.ndarray)
        assert np.array_equal(result, np.array([4., 10., 18.]))

    for lhs in ([4, 5, 6], d2):
        result = lhs * d1
        assert isinstance(result, np.ndarray)
        assert np.array_equal(result, np.array([4., 10., 18.]))


def test_truediv():
    d1 = Dataset(name='/d1', parent=None, read_only=True, data=[1, 2, 1])
    d2 = Dataset(name='/d2', parent=None, read_only=True, data=[4, 4, 10])

    for rhs in ([4., 4., 10.], d2):
        result = d1 / rhs
        assert isinstance(result, np.ndarray)
        assert np.array_equal(result, np.array([0.25, 0.5, 0.1]))

    for lhs in ([4., 4., 10.], d2):
        result = lhs / d1
        assert isinstance(result, np.ndarray)
        assert np.array_equal(result, np.array([4., 2., 10.]))


def test_floordiv():
    d1 = Dataset(name='/d1', parent=None, read_only=True, data=[1e3, 1e4, 1e5])
    d2 = Dataset(name='/d2', parent=None, read_only=True, data=[1e2, 1e3, 1e4])

    for rhs in ([1e2, 1e3, 1e4], d2):
        result = d1 // rhs
        assert isinstance(result, np.ndarray)
        assert np.array_equal(result, np.array([10., 10., 10.]))

    for lhs in ([1e2, 1e3, 1e4], d2):
        result = lhs // d1
        assert isinstance(result, np.ndarray)
        assert np.array_equal(result, np.array([0., 0., 0.]))


def test_pow():
    d1 = Dataset(name='/d1', parent=None, read_only=True, data=[1, 2, 3])
    d2 = Dataset(name='/d2', parent=None, read_only=True, data=[4, 5, 6])

    result = d1 ** 3
    assert isinstance(result, np.ndarray)
    assert np.array_equal(result, np.array([1., 8., 27.]))

    result = pow(d1, 3)
    assert isinstance(result, np.ndarray)
    assert np.array_equal(result, np.array([1., 8., 27.]))

    result = 3 ** d1
    assert isinstance(result, np.ndarray)
    assert np.array_equal(result, np.array([3., 9., 27.]))

    result = pow(3, d1)
    assert isinstance(result, np.ndarray)
    assert np.array_equal(result, np.array([3., 9., 27.]))

    for rhs in ([4., 5., 6.], d2):
        result = d1 ** rhs
        assert isinstance(result, np.ndarray)
        assert np.array_equal(result, np.array([1., 32., 729.]))

        result = pow(d1, rhs)
        assert isinstance(result, np.ndarray)
        assert np.array_equal(result, np.array([1., 32., 729.]))

    for lhs in ([4., 5., 6.], d2):
        result = lhs ** d1
        assert isinstance(result, np.ndarray)
        assert np.array_equal(result, np.array([4., 25., 216.]))

        result = pow(lhs, d1)
        assert isinstance(result, np.ndarray)
        assert np.array_equal(result, np.array([4., 25., 216.]))


@pytest.mark.skipif(sys.version_info[:2] < (3, 5), reason='the @ operator requires Python 3.5+')
def test_matmul():
    import dataset_matmul
    dataset_matmul.run()


def test_mod():
    d = Dataset(name='/d', parent=None, read_only=True, data=list(range(7)))

    result = d % 5
    assert isinstance(result, np.ndarray)
    assert np.array_equal(result, np.array([0, 1, 2, 3, 4, 0, 1]))

    d1 = Dataset(name='/d1', parent=None, read_only=True, data=[4, 7])
    d2 = Dataset(name='/d2', parent=None, read_only=True, data=[2, 3])

    for rhs in ([2, 3], d2):
        result = d1 % rhs
        assert isinstance(result, np.ndarray)
        assert np.array_equal(result, np.array([0, 1]))

    for lhs in ([2, 3], d2):
        result = lhs % d1
        assert isinstance(result, np.ndarray)
        assert np.array_equal(result, np.array([2, 3]))


def test_divmod():
    d1 = Dataset(name='/d1', parent=None, read_only=True, data=[3, 7, 12, 52, 62])
    d2 = Dataset(name='/d2', parent=None, read_only=True, data=np.arange(1, 6))

    for rhs in ([1, 2, 3, 4, 5], d2):
        div, mod = divmod(d1, rhs)
        assert isinstance(div, np.ndarray)
        assert np.array_equal(div, np.array([3,  3,  4, 13, 12]))
        assert isinstance(mod, np.ndarray)
        assert np.array_equal(mod, np.array([0, 1, 0, 0, 2]))

    for lhs in ([1, 2, 3, 4, 5], d2):
        div, mod = divmod(lhs, d1)
        assert isinstance(div, np.ndarray)
        assert np.array_equal(div, np.array([0, 0, 0, 0, 0]))
        assert isinstance(mod, np.ndarray)
        assert np.array_equal(mod, np.array([1, 2, 3, 4, 5]))

    d = Dataset(name='/d', parent=None, read_only=True, data=np.arange(5))
    div, mod = divmod(d, 3)
    assert isinstance(div, np.ndarray)
    assert np.array_equal(div, np.array([0, 0, 0, 1, 1]))
    assert isinstance(mod, np.ndarray)
    assert np.array_equal(mod, np.array([0, 1, 2, 0, 1]))


def test_lshift():
    d1 = Dataset(name='/d1', parent=None, read_only=True, data=np.array([1, 2, 3, 4, 5]))
    d2 = Dataset(name='/d2', parent=None, read_only=True, data=np.array([3, 7, 11, 15, 19]))

    result = d1 << 1
    assert isinstance(result, np.ndarray)
    assert np.array_equal(result, np.array([2, 4, 6, 8, 10]))

    result = 1 << d1
    assert isinstance(result, np.ndarray)
    assert np.array_equal(result, np.array([2, 4, 8, 16, 32]))

    for rhs in ([3, 7, 11, 15, 19], d2):
        result = d1 << rhs
        assert isinstance(result, np.ndarray)
        assert np.array_equal(result, np.array([8, 256, 6144, 131072, 2621440]))

    for lhs in ([3, 7, 11, 15, 19], d2):
        result = lhs << d1
        assert isinstance(result, np.ndarray)
        assert np.array_equal(result, np.array([6, 28, 88, 240, 608]))


def test_rshift():
    d1 = Dataset(name='/d1', parent=None, read_only=True, data=np.array([1, 2, 3, 4, 5]))
    d2 = Dataset(name='/d2', parent=None, read_only=True, data=np.array([3, 7, 12, 52, 62]))

    result = d1 >> 10
    assert isinstance(result, np.ndarray)
    assert np.array_equal(result, np.array([0, 0, 0, 0, 0]))

    result = 10 >> d1
    assert isinstance(result, np.ndarray)
    assert np.array_equal(result, np.array([5, 2, 1, 0, 0]))

    for rhs in ([3, 7, 12, 52, 62], d2):
        result = d1 >> rhs
        assert isinstance(result, np.ndarray)
        assert np.array_equal(result, np.array([0, 0, 0, 0, 0]))

    for lhs in ([3, 7, 12, 52, 62], d2):
        result = lhs >> d1
        assert isinstance(result, np.ndarray)
        assert np.array_equal(result, np.array([1, 1, 1, 3, 1]))


def test_and():
    d1 = Dataset(name='/d1', parent=None, read_only=True, data=np.arange(9))
    d2 = Dataset(name='/d2', parent=None, read_only=True, data=np.arange(10, 19))

    for rhs in ([10, 11, 12, 13, 14, 15, 16, 17, 18], d2):
        result = d1 & rhs
        assert isinstance(result, np.ndarray)
        assert np.array_equal(result, np.array([0, 1, 0, 1, 4, 5, 0, 1, 0]))

    for lhs in ([10, 11, 12, 13, 14, 15, 16, 17, 18], d2):
        result = lhs & d1
        assert isinstance(result, np.ndarray)
        assert np.array_equal(result, np.array([0, 1, 0, 1, 4, 5, 0, 1, 0]))


def test_xor():
    d1 = Dataset(name='/d1', parent=None, read_only=True, data=np.arange(9))
    d2 = Dataset(name='/d2', parent=None, read_only=True, data=np.arange(10, 19))

    for rhs in ([10, 11, 12, 13, 14, 15, 16, 17, 18], d2):
        result = d1 ^ rhs
        assert isinstance(result, np.ndarray)
        assert np.array_equal(result, np.array([10, 10, 14, 14, 10, 10, 22, 22, 26]))

    for lhs in ([10, 11, 12, 13, 14, 15, 16, 17, 18], d2):
        result = lhs ^ d1
        assert isinstance(result, np.ndarray)
        assert np.array_equal(result, np.array([10, 10, 14, 14, 10, 10, 22, 22, 26]))


def test_or():
    d1 = Dataset(name='/d1', parent=None, read_only=True, data=np.arange(9))
    d2 = Dataset(name='/d2', parent=None, read_only=True, data=np.arange(10, 19))

    for rhs in ([10, 11, 12, 13, 14, 15, 16, 17, 18], d2):
        result = d1 | rhs
        assert isinstance(result, np.ndarray)
        assert np.array_equal(result, np.array([10, 11, 14, 15, 14, 15, 22, 23, 26]))

    for lhs in ([10, 11, 12, 13, 14, 15, 16, 17, 18], d2):
        result = lhs | d1
        assert isinstance(result, np.ndarray)
        assert np.array_equal(result, np.array([10, 11, 14, 15, 14, 15, 22, 23, 26]))


def test_neg():
    # unary "-"

    d1 = Dataset(name='/d1', parent=None, read_only=True, data=[1, 2, 3])
    d2 = Dataset(name='/d2', parent=None, read_only=True, data=[4, 5, 6])

    for rhs in [[4, 5, 6], d2]:
        result = -d1 + rhs
        assert isinstance(result, np.ndarray)
        assert np.array_equal(result, np.array([3, 3, 3]))


def test_pos():
    # unary "+"

    d1 = Dataset(name='/d1', parent=None, read_only=True, data=[1, 2, 3])
    d2 = Dataset(name='/d2', parent=None, read_only=True, data=[4, 5, 6])

    for rhs in [[4, 5, 6], d2]:
        result = +d1 - rhs
        assert isinstance(result, np.ndarray)
        assert np.array_equal(result, np.array([-3, -3, -3]))


def test_abs():
    # unary "abs()"

    d = Dataset(name='/d', parent=None, read_only=True, data=[1, -2, 3, -4])

    result = abs(d)
    assert isinstance(result, np.ndarray)
    assert np.array_equal(result, np.array([1, 2, 3, 4]))


def test_invert():
    # unary "~"

    d = Dataset(name='/d', parent=None, read_only=True, data=np.array([1, -2, 3, -4]))

    result = ~d
    assert isinstance(result, np.ndarray)
    assert np.array_equal(result, np.array([-2, 1, -4, 3]))


def test_assignments():
    d = Dataset(name='/d', parent=None, read_only=True, data=np.array([1, 2, 3]))
    d += 1
    assert isinstance(d, np.ndarray)
    assert np.array_equal(d, np.array([2, 3, 4]))

    d = Dataset(name='/d', parent=None, read_only=True, data=np.array([1, 2, 3]))
    d -= 1
    assert isinstance(d, np.ndarray)
    assert np.array_equal(d, np.array([0, 1, 2]))

    d = Dataset(name='/d', parent=None, read_only=True, data=np.array([1, 2, 3]))
    d *= 10
    assert isinstance(d, np.ndarray)
    assert np.array_equal(d, np.array([10, 20, 30]))

    d = Dataset(name='/d', parent=None, read_only=True, data=np.array([10, 20, 30]))
    d /= 10
    assert isinstance(d, np.ndarray)
    assert np.array_equal(d, np.array([1, 2, 3]))

    d = Dataset(name='/d', parent=None, read_only=True, data=np.array([10, 20, 30]))
    d //= 5
    assert isinstance(d, np.ndarray)
    assert np.array_equal(d, np.array([2, 4, 6]))

    d = Dataset(name='/d', parent=None, read_only=True, data=np.array([10, 20, 30]))
    d %= 15
    assert isinstance(d, np.ndarray)
    assert np.array_equal(d, np.array([10, 5, 0]))

    d = Dataset(name='/d', parent=None, read_only=True, data=np.array([1, 2, 3]))
    d **= 3
    assert isinstance(d, np.ndarray)
    assert np.array_equal(d, np.array([1, 8, 27]))

    d = Dataset(name='/d', parent=None, read_only=True, data=np.array([1, 2, 3]))
    d <<= 3
    assert isinstance(d, np.ndarray)
    assert np.array_equal(d, np.array([8, 16, 24]))

    d = Dataset(name='/d', parent=None, read_only=True, data=np.array([10, 20, 30]))
    d >>= 2
    assert isinstance(d, np.ndarray)
    assert np.array_equal(d, np.array([2, 5, 7]))

    d = Dataset(name='/d', parent=None, read_only=True, data=np.array([1, 2, 3]))
    d &= 2
    assert isinstance(d, np.ndarray)
    assert np.array_equal(d, np.array([0, 2, 2]))

    d = Dataset(name='/d', parent=None, read_only=True, data=np.array([1, 2, 3]))
    d ^= 2
    assert isinstance(d, np.ndarray)
    assert np.array_equal(d, np.array([3, 0, 1]))

    d = Dataset(name='/d', parent=None, read_only=True, data=np.array([1, 2, 3]))
    d |= 2
    assert isinstance(d, np.ndarray)
    assert np.array_equal(d, np.array([3, 2, 3]))


def test_numpy_function():
    # np.xxx() is also valid syntax with a Dataset

    array = np.array([1, 2, 3])
    d1 = Dataset(name='/d1', parent=None, read_only=True, data=[1, 2, 3])

    cos = np.cos(d1)
    assert isinstance(cos, np.ndarray)
    assert np.array_equal(cos, np.cos(array))

    sqrt = np.sqrt(d1)
    assert isinstance(sqrt, np.ndarray)
    assert np.array_equal(sqrt, np.sqrt(array))

    abs_ = np.abs(d1)
    assert isinstance(abs_, np.ndarray)
    assert np.array_equal(abs_, np.abs(array))

    max_ = np.max(d1)
    assert isinstance(max_, float)
    assert np.array_equal(max_, np.max(array))



"""
Constants used by MSL-IO.
"""
import os
import sys

IS_PYTHON2 = sys.version_info.major == 2
""":class:`bool`: Whether Python 2 is being used."""

IS_PYTHON3 = sys.version_info.major == 3
""":class:`bool`: Whether Python 3 is being used."""

# If this module is run via "sudo python" on a Raspberry Pi the value of
# os.path.expanduser('~') becomes '/root' instead of '/home/pi'. On Linux using
# "sudo python" keeps os.path.expanduser('~') as /home/<username> and running this
# module in an elevated command prompt on Windows keeps os.path.expanduser('~')
# as C:\\Users\\<username>. Therefore defining USER_DIR in the following way keeps
# things more consistent across more platforms.
USER_DIR = os.path.expanduser('~'+os.getenv('SUDO_USER', ''))

HOME_DIR = os.getenv('MSL_IO_HOME', os.path.join(USER_DIR, '.msl', 'io'))
""":class:`str`: The default directory where all files that are used by MSL-IO are located.

Can be overwritten by specifying a ``MSL_IO_HOME`` environment variable.
"""



# -*- coding: utf-8 -*-
import os
import tempfile

import numpy as np
import pytest

try:
    import h5py
except ImportError:
    h5py = None

from helper import read_sample
from helper import roots_equal
from msl.io import HDF5Writer
from msl.io import JSONWriter
from msl.io import read
from msl.io.readers import JSONReader


def test_read_write_convert():
    root1 = read_sample('json_sample.json')

    # write as JSON then read
    writer = JSONWriter(tempfile.gettempdir() + '/msl-json-writer-temp.json')
    writer.write(root=root1, mode='w')
    root2 = read(writer.file)
    assert root2.file == writer.file
    assert roots_equal(root1, root2)
    os.remove(writer.file)

    # convert to HDF5 then back to JSON
    hdf5_writer = HDF5Writer(tempfile.gettempdir() + '/msl-hdf5-writer-temp.h5')
    # HDF5 does not support "null". So, reload the JSON file and pop the metadata
    # that contain "null" before writing the HDF5 file
    temp = read_sample('json_sample.json')
    assert roots_equal(root1, temp)
    temp.read_only = False
    assert temp.metadata.pop('null') is None
    assert 'null' not in temp.metadata
    array_mixed_null = temp.metadata.pop('array_mixed_null')
    assert isinstance(array_mixed_null, np.ndarray)
    assert np.array_equal(array_mixed_null, ['a', False, 5, 72.3, 'hey', None])
    assert 'array_mixed_null' not in temp.metadata
    if h5py is not None:
        hdf5_writer.write(root=temp, mode='w')
        root_hdf5 = read(hdf5_writer.file)
        os.remove(hdf5_writer.file)
        writer2 = JSONWriter(tempfile.gettempdir() + '/msl-json-writer-temp2.json')
        writer2.write(root=root_hdf5, mode='w')
        root3 = read(writer2.file)
        assert root3.file == writer2.file
        os.remove(writer2.file)

    roots = [root1, root2]
    if h5py is not None:
        roots.append(root3)

    for root in roots:
        assert isinstance(root, JSONReader)

        assert len(list(root.groups())) == 4
        assert len(list(root.a.groups())) == 2
        assert len(list(root.a.b.groups())) == 1
        assert len(list(root.a.b.c.groups())) == 0

        assert len(list(root.datasets())) == 2
        assert len(list(root.a.datasets())) == 1
        assert len(list(root.a.b.datasets())) == 1
        assert len(list(root.a.b.c.datasets())) == 0

        # HDF5 does not support NULL type values
        # also, "array_mixed" gets converted to be all strings by h5py
        if h5py is not None and root is root3:
            assert len(root.metadata) == 9
            # use tolist() not np.array_equal
            assert root.metadata.array_mixed.tolist() == ['True', '-5', '0.002345', 'something', '49.1871524']
        else:
            assert len(root.metadata) == 11
            assert root.metadata.null is None
            # use tolist() not np.array_equal
            assert root.metadata.array_mixed_null.tolist() == ['a', False, 5, 72.3, 'hey', None]
            assert root.metadata.array_mixed.tolist() == [True, -5, 0.002345, 'something', 49.1871524]

        assert root.metadata.foo == 'bar'
        assert root.metadata.boolean is True
        assert root.metadata['integer'] == -99
        assert root.metadata['float'] == 33.33e3
        assert root.metadata['empty_list'].size == 0
        # use tolist() not np.array_equal
        assert root.metadata.array_strings.tolist() == ['aaa', 'bbb', 'ccc', 'ddd', 'eee']
        assert root.metadata.array_var_strings.tolist() == ['a', 'ab', 'abc', 'abcd', 'abcde']
        assert root.metadata.array_numbers.tolist() == [1, 2.3, 4, -4e99]

        # make sure the Metadata values are read only
        with pytest.raises(ValueError):
            root.metadata['new_key'] = 'new value'
        with pytest.raises(ValueError):
            del root.metadata.foo
        with pytest.raises(ValueError):
            root.metadata.boolean = False
        with pytest.raises(ValueError):
            root.metadata.array_strings[0] = 'new string'
        with pytest.raises(ValueError):
            del root.metadata.array_var_strings
        with pytest.raises(ValueError):
            root.metadata.array_numbers[:] = [-9, -8, -7, -6]

        root.read_only = False

        # can now modify the Metadata
        del root.metadata.foo
        assert 'foo' not in root.metadata
        root.metadata.boolean = False
        assert root.metadata.boolean is False
        root.metadata.array_strings[0] = 'new string'
        # use tolist() not np.array_equal
        assert root.metadata.array_strings.tolist() == ['new string', 'bbb', 'ccc', 'ddd', 'eee']
        root.metadata.array_numbers[::2] = [-9, -8.8]
        assert root.metadata.array_numbers.tolist() == [-9, 2.3, -8.8, -4e99]
        assert 'new_key' not in root.metadata
        root.metadata.new_key = 'new value'
        assert root.metadata.new_key == 'new value'

        root.read_only = True

        # make sure the Metadata values are read only again
        with pytest.raises(ValueError):
            del root.metadata.new_key
        with pytest.raises(ValueError):
            root.metadata.boolean = True
        with pytest.raises(ValueError):
            root.metadata.array_strings[-1] = 'another string'
        with pytest.raises(ValueError):
            del root.metadata.array_var_strings
        with pytest.raises(ValueError):
            root.metadata.array_numbers[:] = [11, 22, 33, 44]

        assert 'conditions' in root
        assert 'my_data' in root
        assert 'a' in root
        assert '/a/b' in root
        assert '/a/b/c' in root
        assert '/a/b/dataset2' in root

        assert root.is_group(root.conditions)
        assert root['conditions'].metadata['temperature'] == 20.0
        assert root.conditions.metadata.humidity == 45

        assert root['my_data'].metadata.meta == 999
        assert root.is_dataset(root['my_data'])
        assert 'dtype' not in root['my_data'].metadata
        assert isinstance(root['my_data'].data, np.ndarray)
        assert root['my_data'].shape == (4, 5)
        assert root['my_data'].dtype.str == '<i4'
        assert np.array_equal(root['my_data'], np.arange(20).reshape(4, 5))

        assert len(root.a.metadata) == 0

        b = root.a.b
        assert len(b.metadata) == 2
        assert b.metadata.apple == 'orange'
        assert b.metadata.pear == 'banana'

        c = root.a.b.c
        assert root.is_group(c)
        assert len(c.metadata) == 0

        dset = root.a.b.dataset2
        assert root.is_dataset(dset)
        assert len(dset.metadata) == 1
        # use tolist() not np.array_equal
        assert dset.metadata.fibonacci.tolist() == [1, 1, 2, 3, 5, 8, 13, 21, 34, 55]
        assert dset.shape == (2,)
        assert dset.dtype.names == ('a', 'b', 'c', 'd', 'e')
        assert dset['a'].dtype == object
        assert dset['b'].dtype == object
        assert dset['c'].dtype == float
        assert dset['d'].dtype == float
        assert dset['e'].dtype == np.int32
        assert dset['a'].tolist() == ['100', '100s']
        assert dset['b'].tolist() == ['100s', '50+50s']
        assert dset['c'].tolist() == [0.000640283, -0.000192765]
        assert dset['d'].tolist() == [0.0, 11.071]
        assert dset['e'].tolist() == [8, 9]

        # make sure the ndarray's are read only
        with pytest.raises(ValueError):
            dset.metadata.fibonacci[4] = -9
        with pytest.raises(ValueError):
            dset['e'] = [-1, 0]

        root.read_only = False

        dset.metadata.fibonacci[4] = -9
        assert dset.metadata.fibonacci.tolist() == [1, 1, 2, 3, -9, 8, 13, 21, 34, 55]
        dset['e'] = [-1, 0]
        assert dset['e'].tolist() == [-1, 0]

        root.read_only = True

        # make sure the ndarray's are read only again
        with pytest.raises(ValueError):
            dset.metadata.fibonacci[4] = 0
        with pytest.raises(ValueError):
            dset['a'][0] = 'foo'


def test_raises():
    root = read_sample('json_sample.json')

    writer = JSONWriter()
    assert writer.file is None

    # no file was specified
    with pytest.raises(ValueError, match=r'must specify a file'):
        writer.write(root=root)

    # root must be a Root
    with pytest.raises(TypeError, match=r'Root'):
        writer.write(file='whatever', root=list(root.datasets())[0])
    with pytest.raises(TypeError, match=r'Root'):
        writer.write(file='whatever', root=list(root.groups())[0])
    with pytest.raises(TypeError, match=r'Root'):
        writer.write(file='whatever', root='Root')

    # cannot overwrite a file by default
    file = tempfile.gettempdir() + '/msl-json-writer-temp.json'
    with open(file, mode='wt') as fp:
        fp.write('Hi')
    with pytest.raises(OSError, match=r'File exists'):
        writer.write(file=file, root=root)
    with pytest.raises(OSError, match=r'File exists'):
        writer.write(file=file, root=root, mode='x')

    # invalid mode
    for m in ['r', 'z']:
        with pytest.raises(ValueError, match=r'([i|I]nvalid\s)?mode'):
            writer.write(file=file, root=root, mode=m)

    # r+ is a valid mode, but the file must already exist
    with pytest.raises((IOError, OSError), match=r'No such file or directory'):
        writer.write(file='does_not.exist', root=root, mode='r+')

    # by specifying the mode one can overwrite a file
    writer.write(file=file, root=root, mode='w')
    assert roots_equal(root, read(file))
    writer.write(file=file, root=root, mode='w+')
    assert roots_equal(root, read(file))
    writer.write(file=file, root=root, mode='r+')
    assert roots_equal(root, read(file))
    writer.write(file=file, root=root, mode='a')
    assert roots_equal(root, read(file))
    os.remove(file)


def test_pretty_printing():
    root = read_sample('json_sample.json')
    root.read_only = False

    root.create_dataset('aaa', data=np.ones((3, 3, 3)))

    w = JSONWriter(tempfile.gettempdir() + '/msl-json-writer-temp.json')
    w.save(root=root, mode='w', sort_keys=True, separators=(', ', ': '))

    expected = """#File created with: MSL JSONWriter version 1.0
{
  "a": {
    "b": {
      "apple": "orange",
      "c": {},
      "dataset2": {
        "data": [
          ["100", "100s", 0.000640283, 0.0, 8],
          ["100s", "50+50s", -0.000192765, 11.071, 9]
        ],
        "dtype": [
          ["a", "object"],
          ["b", "object"],
          ["c", "float64"],
          ["d", "float64"],
          ["e", "int32"]
        ],
        "fibonacci": [1, 1, 2, 3, 5, 8, 13, 21, 34, 55]
      },
      "pear": "banana"
    }
  },
  "aaa": {
    "data": [
      [
        [1.0, 1.0, 1.0],
        [1.0, 1.0, 1.0],
        [1.0, 1.0, 1.0]
      ],
      [
        [1.0, 1.0, 1.0],
        [1.0, 1.0, 1.0],
        [1.0, 1.0, 1.0]
      ],
      [
        [1.0, 1.0, 1.0],
        [1.0, 1.0, 1.0],
        [1.0, 1.0, 1.0]
      ]
    ],
    "dtype": "<f8"
  },
  "array_mixed": [true, -5, 0.002345, "something", 49.1871524],
  "array_mixed_null": ["a", false, 5, 72.3, "hey", null],
  "array_numbers": [1, 2.3, 4, -4e+99],
  "array_strings": ["aaa", "bbb", "ccc", "ddd", "eee"],
  "array_var_strings": ["a", "ab", "abc", "abcd", "abcde"],
  "boolean": true,
  "conditions": {
    "humidity": 45,
    "temperature": 20.0
  },
  "empty_list": [],
  "float": 33330.0,
  "foo": "bar",
  "integer": -99,
  "my_data": {
    "data": [
      [0, 1, 2, 3, 4],
      [5, 6, 7, 8, 9],
      [10, 11, 12, 13, 14],
      [15, 16, 17, 18, 19]
    ],
    "dtype": "<i4",
    "meta": 999
  },
  "null": null
}
""".splitlines()

    with open(w.file, mode='rt') as fp:
        written = [line.rstrip() for line in fp.read().splitlines()]

    assert len(expected) == len(written)
    for i in range(len(expected)):
        assert expected[i] == written[i]

    # make sure that we can still read the file
    root = read_sample(w.file)

    # change the indentation to be 0
    w.save(root=root, mode='w', sort_keys=True, indent=0, separators=(', ', ': '))
    with open(w.file, mode='rt') as fp:
        written = [line.rstrip() for line in fp.read().splitlines()]
    assert len(expected) == len(written)
    for i in range(len(expected)):
        assert expected[i].lstrip() == written[i]

    # change the indentation to be None
    w.save(root=root, mode='w', sort_keys=True, indent=None, separators=(',', ':'))
    with open(w.file, mode='rt') as fp:
        written = fp.read().splitlines()
    assert len(written) == 2
    assert written[0] == '#File created with: MSL JSONWriter version 1.0'
    assert written[1].startswith('{"a":{"b":{"apple":')
    assert written[1].endswith('},"null":null}')

    os.remove(w.file)


def test_unicode():
    def do_asserts(r):
        assert len(r.metadata) == 2
        assert u'μ' in r.metadata
        assert u'\u03bc' in r.metadata
        assert 'unit' in r.metadata
        assert r.metadata[u'μ'] == 'micro'
        assert r.metadata[u'\u03bc'] == 'micro'
        assert r.metadata.unit == u'°C'
        assert r.metadata.unit == u'\xb0C'

    root = read_sample(u'uñicödé.json')
    do_asserts(root)

    for b in [False, True]:
        writer = JSONWriter(tempfile.gettempdir() + '/msl-json-writer-temp.json')
        writer.save(root=root, ensure_ascii=b, mode='w')

        root2 = read(writer.file)
        do_asserts(root2)
        assert roots_equal(root, root2)

        os.remove(writer.file)
